;FILE: MT63ASC.ASM, Last edition: 15-AUG-1997
;64-tone, 10 baud, DBPSK HF modem with ASCII coding
;(c) 1997 Pawel Jalocha, SP9VRC,
;e-mail: Pawel.Jalocha@cern.ch
;
;This software is not to be used for purposes other than amateur radio
;without a written permision of the author.
;Usage for profit or in commercial/military products is explicitly prohibited.
;
;Note: the frequency offset shown by the Rx quality report is NOT correct.
;      I shall repair it when I am back from holidays.
;The general features of the modem:
; - the modem sends 64 tones spaced by 15.625 Hz (8000 Hz/5/512*5)
; - the modem audio occupies the range from 500 to 1500 Hz
; - tones are modulated with differential bipolar phase shift keying (DBPSK)
; - modulation rate is 10 baud per tone thus the raw data rate is 640 bps
; - the user data are 7-bit ASCII characters encoded with Walsh functions:
;   a 7-bit character is enocoded into a block of 64 bits. Up to 16 bits
;   out of 16 can be corrupted before the character itself becomes corrupted
; - each block is spread over 32 symbols (3.2 seconds) in time
;   and over all tones to minimize the influence of time or frequency
;   localized interferences
; - the signal capture logic is able to catch the lock at any moment within
;   a transmition and the frequency tolerance is about +/- 50 Hz
; - the tracking logic can track timing and frequency shift infinitely
; - in this version the demodulator is purely phase-type so it doesn't care
;   for any amplitude variations across time or frequency
; - the demodulator supplies soft decisions in the form: cos(phase_diff)
;   where phase_diff is the difference in phases between succesive samples
; - if you want to capture a multitone signal and tolerate the absence
;   of some tones you are going to face the frequency ambiguity - it is
;   resolved here by trying all possible combinations and selecting the one
;   where the Walsh function decoder supplies least error.
;
;Operating the modem:
; - you need an ASCII terminal, you may find it more confortable to
;   have separate transmit and receive screen
;Transmition:
; - whatever you type will be accumulated in a buffer (TxCharBuffLen
;   characters long)
; - when you type Ctrl-T the modem start the transmition: it sends
;   the text defined as TxStartSeq, then the character you typed in
;   and eventually more characters you typed in the mean time.
;   If there is nothing to be transmitted the modem will send NUL's.
;   If TxReport is set the modem will echo every character it transmits.
; - when you type Ctrl-R the modem sends out what is left in the buffer,
;   then the TxStopSeq and then it stops and releases the PTT.
; - typing Ctrl-A during the transmition will immediately shut down
;   the transmitter
;Reception:
; - when the modem locks onto a transmition it will put out a message
;   (if the RxReport is set) and then it will send the decoded
;   characters.
; - when the carrier is lost the modem will stop decoding and
;   it will put out a message (if RxReport is set)
; - when you type Ctrl-D while the modem is decoding the carrier
;   will be dropped and a new acquisition attempt will start
; - when you type Ctrl-Q while decoding the modem will put out
;   a signal quality report telling the frequency offset and S/N
;   seen by the BPSK demodulator and the Walsh function FEC decoder
;Other notes:
; - there are lot of options (just below) which you might want to change
; - not all possible combinations of every option have been tested
;   even some of the options have never been tried :-)
; - I didn't try to compile this code for the DSPCARD4 but it should work
;   there - you may need to rearrange some tables but that's all.
;   The CPU consumption is 30% by the receiver and 10% for the transmitter
;   on my 40 MHz EVM56k thus the 27 MHz DSPCARD4 should do it.
; - if you (or if you don't) manage to make actuall QSOs with this code
;   please send me a report with short description of the performance
;   and the radio conditions

        page 132
        nolist
	include 'leonid'
	list

        scsjmp short    ;force short jumps for .if/.endi, .while/.endw, etc.
        opt nops        ;do not pack strings (make one character per word)

        title '64-tone, 10 baud DBPSK HF modem by SP9VRC'

;*************************************************************************
;compile-time constants

;the EVM56K is now defined in leonid.asm
;EVM56K  equ 0           ;0 => DSPCARD4
                        ;1 => EVM56002

SampleFreq equ 8000.0           ;Sampling rate of the CODEC
                                ;you can change it to 9600 Hz and the whole
                                ;modem will scale up by 9600/8000
                                ;without the need to touch anything else.
DecimateRatio equ 5             ;decimation ratio
                                ;the maximum bandwidth of the anti-alias filter:
                                ;sampled bandwidth = 8000/5 = 1600 Hz
AliasFilterLen equ 96           ;the length of the anti-alias FIR filter
WindowLen       equ 512         ;sliding-window length
WindowLenLog    equ 9           ;and its 2-base logarythm
                                ; => FT freq. bin = 8000/5/512 = 3.125 Hz
SymbolSepar     equ 160         ;the time-spacing of our symbols
                                ; => symbol rate = 8000/5/160 = 10 baud
                                ;The SymbolSepar should be a multiply of 2
                                ;and should not be bigger than WindowLen/2.
                                ;If you change SymbolSepar you should change
                                ;the shape of the symbol too (symbol.dat),
                                ;otherwise you will get significant ISI.
TxSliceSepar    equ SymbolSepar/4 ;default times between the FFT slices
RxSliceSepar    equ SymbolSepar/4 ;not to be changed.
FirstDataCarr   equ 162         ;[FFT bins] ;first data carrier at 162*3.125 Hz
                                ;Changing FirstDataCarr should be coordinated
                                ;with modifications of the anti-alias filter
DataCarrSepar   equ 5 ;[FFT bins] ;data carriers separated by 5*3.125 Hz
                                ;(a very fixed parameter)
DataCarriers    equ 64 ;[integer] ;64 data carriers
DataCarriersLog equ 6           ;This code needs powers of 2.
                                ; => middle carrier is at about 1000 Hz
                                ; => audio spectra is 1000+/-500 Hz
BitsPerBlock    equ DataCarriers ;one bit per carrier forms a block
WordsPerBlock   equ (BitsPerBlock-1)/24+1 ;number of machine words (24 bits)
                                ;per transmitted block (64 bits)
DataInterleave  equ DataCarriers/2 ;Interleave level. This code is fixed
                                ;at 32 - we can't do more due to little RAM.
TxMinIdle       equ 50 ;[symbols]  minimal idling time
                                ;This is to prevent the new transmition
                                ;to immediately follow the previous one.
                                ;I'm not sure if this feature is usefull here.
TxSyncLen       equ 40 ;[symbols] the length of the initial sequence
                                ; 32/10 baud = 3.2 sec
                                ;Modify the TxStartCharSeq after TxSyncLen
TxStopLen       equ 19 ;[symbols] the length of the terminating sequence
                                ;modify the TxStopCharSeq after TxStopLen
TxJamLen        equ 16  ;[symbols] the time for which the jamming sequence
                                ;is transmitted (for faster DCD turn-off).
TxDataAmpl equ 1.0/32           ;amplitude for symbols to be put into
                                ;the FFT bins at the transmitter.
                                ;If you put too much the FFT may overflow (!)
                                ;and you will transmit a mess.
                                ;we only transmform half of the symbols
                                ;at one FFT thus 1/32 should be safe.
                                ;By looking at the SPY_TxAudio one can see
                                ;that we do not use the full output range
                                ;and we could still go up by a factor of 4
                                ;by enlarging TxDataAmpl or by scaling up
                                ;the output interpolation filter.
TxAttenuate     equ  0.0  ;[dB] attenuate the output CODEC's signal.
                                ;Set this between 0.0 and 90.0
                                ;to accomodate the output level for the
                                ;modulator input of your tranceiver.
                                ;Experiments indicate that when you attenuate
                                ;here the signal is noisy and distorted
                                ;thus it is recomended to use an external
                                ;attenuator.
TxCharBuffLen   equ 256         ;Tx type-ahead buffer size
TxReport        equ 1           ;Report transmition start/stop and as well
                                ;echo the characters while they are being
                                ;transmitted. You may not like this unless
                                ;your terminal has separate transmit
                                ;and receive screens
RxSyncScanMargin equ 16 ;[FFT bins] ;the receiver scans that wide range
                                ;to look for data tones.
                                ;Unfortunately we can lock and be shifted
                                ;by one or more carriers. This problem
                                ;we will try to resolve after the lock.
RxSyncScanFirst   equ FirstDataCarr-RxSyncScanMargin
RxSyncScanLen     equ (DataCarriers-1)*DataCarrSepar+1+2*RxSyncScanMargin
RxSyncPhaseDevLen equ DataCarrSepar+2*RxSyncScanMargin+1
RxSyncAverWeight  equ 1.0/16    ;Averageing weight for accumulating
                                ;the lock confidence
RxSyncCatchThres equ 0.29       ;Threshold for catching the lock
                                ;Lower threshold => more S/N required for lock
                                ;Higher threshold => locks on weaker signals
                                ;but may sporadically catch on noise.
                                ;The critical level where you get
                                ;fake lock on pure noise is 0.33
                                ;Setting smaller (like 1/32.0) RxSyncAverWeight
                                ;allows you to get closer to the noise limit
                                ;but prolong the synchronization.
RxSyncAvoidPTT  equ 0           ;avoid collecting the sync. statistics
                                ;while the PTT is on. This can help you
                                ;when you are working in half-duplex
                                ;(most people do) and you have audio feedback
                                ;through the tranceiver
RxDataScanMargin    equ 4       ;tolerance for locking off by one or more
                                ;carriers. Actuall tolerance is twice this
                                ;in terms of DataCarrSepar.
RxDataScanLen   equ DataCarriers+2*2*RxDataScanMargin
                                ;that many FFT bins separated by DataCarrSepar
                                ;are decoded during the data phase
RxDataSampleLen equ DataCarriers/2+2*RxDataScanMargin
                                ;Data (soft) sample pipe width
RxDataPhaseDevLen equ 2+2*2*RxDataScanMargin
                                ;That many posible carrier position we will
                                ;trace for phase deviation
RxDataAverWeight  equ 1.0/16    ;Weight for averaging tracking quantities
                                ;during the data phase
RxDataDropThres   equ 0.30      ;Threshold for droping the lock.
                                ;Pure noise level is 0.33
                                ;It is wise to put RxDataDropThres a bit
                                ;above the RxSyncCatchThres
RxDataCorrAverShift   equ 0     ;correct average phase shift for a symbol
                                ;this may help if there are sudden frequency
                                ;drifts
RxDataFreqTrackWeight equ 1.0/24 ;1.0/16 makes a 20% overshoot
RxCharSelThres  equ 0.00        ;selection threshold for decoding character
                                ;if the soft decoder noise on a character
                                ;is higher then the character will be
RxBadCharFill   equ '_'         ;replaced by RxBadCharFill
                                ;My experiments with this feature
                                ;didn't show any advantage :-)
RxGain          equ 10.0 ;[dB] Rx CODEC's input gain
                                ;set this between 0 and 22.5 dB depending on
                                ;the audio level from your tranceiver
RxAGC           equ 1           ;enable the automatic gain adjustment
                                ;for the CODEC's audio input.
                                ;ideally you should avoid the AGC and set
                                ;RxGain for best audio level match.
                                ;Note, that the AGC may get fooled up by
                                ;for example long periods of total silence,
                                ;thus you should rather keep the quelch (if any)
                                ;open all the time.
RxAGC_MSfollow  equ 1.0/8.0     ;update weight per quarter-symbol
                                ;for the input audio Mean Square tracking
RxAGChold       equ 100 ;[symbol/4]  AGC peak hold time
RxAGCfall       equ 10  ;[symbol/4]  AGC falling period/1.5 dB step
RxAGCminMS      equ 0.02        ;min. and max. tolerable audio Mean Square level
RxAGCmaxMS      equ 0.05        ;must not be too close because CODEC can only
                                ;move in 1.5 dB steps
RxAGCmaxPeak    equ 0.75        ;Max. allowed peak audio level
RxAGCsensePeak  equ 0           ;Take (or don't take) into acount the peak level
                                ;when deciding about the gain changes
                                ;taking the peak into account makes the AGC
                                ;more "jumpy" which is not good in my opinion.
RxAGCavoidDCD   equ 0           ;Disallow gain changes when DCD is on
                                ;so the gain changes do not mess up
                                ;the transmition being received.
RxAGCavoidPTT   equ 0           ;The AGC would ignore the silence on the Rx
                                ;when the PTT is on. Usefull for simplex
                                ;but not good for full duplex tests.
RxReport        equ 1           ;report signal acquisition and loss
                                ;on the terminal
RightChannel    equ 0           ;Use the right CODEC's channel not the left.
                                ;I didn't try this but Timo says it works.
NewBaudRate     equ 0           ;Set a new baud rate on the serial port
                                ;for faster KISS or SPYing
                                ;When 0 the serial port runs at BIOS'es rate.
CodecBuffLen  equ 512   ;Sample buffer length with some safety margin
                        ;but not too much because large CodecBuffLen make extra
                        ;delays in the transmitions (we have to hold the PTT
                        ;until this buffer is flushed) so one can get some
                        ;improvement in the turn-around time by optimizing
                        ;CodecBufFLen.
                        ;CodecBuffLen must not be smaller than
                        ;AliasFilterLen+DecimateRatio*SymbolSepar
                        ;Also: smaller CodecBuffLen makes the code intolerant
                        ;to sudden processing demand (like to process
                        ;a bigger FEC block or to send out a long message).

SPY      equ 0  ; SPY=0 => terminal mode
                ; SPY=1 => terminal code excluded, selected debug code installed
SPY_BlockLen equ 512    ;how many words to follow each SPY request,
                        ;originally 512 but can be changed here.
                        ;If you are spying on very fast data don't put more
                        ;than what the serial output buffer can hold.
;Select one of the SPY_ing options below
;Rx sensitive points:
SPY_RxAudio     equ 0   ;Rx audio just after CODEC conversion
SPY_RxIQ        equ 0   ;I/Q after the Hilbert transformer and Doppler corrector
SPY_RxCarFreq   equ 0   ;the doppler correction
SPY_RxSliceSepar equ 0  ;the distance between FFT slices
SPY_RxFFT       equ 0   ;spectral complex values just after FFT
SPY_RxFFTpower  equ 0   ;spectral power (I*I+Q*Q) values just after FFT
SPY_PeakMS      equ 0   ;signal's peak and MS seen by the AGC
SPY_RxCodecPtr  equ 0   ;the Rx CODEC pointer
SPY_RxStateRout equ 0   ;the state routine being in effect
SPY_RxCorrFFT   equ 0   ;FFT data after phase correction
SPY_RxSelFFTbin equ -162 ;spy a selected FFT bin (negative disables)
SPY_RxSelFFTphase equ -162 ;spy the phase of a selected FFT bin

SPY_RxSyncPhases   equ 0   ;all differential phases for the sync. hunting phase
SPY_RxSyncSelDiffPhase equ -2 ;selected differential phase (negative disables)
SPY_RxSyncPhaseDev     equ 0
SPY_RxSyncMinPhases    equ 0
SPY_RxSyncScanMin      equ 0   ;the best fit to the carrier pattern
;SPY_RxSyncTiming      equ 0

SPY_RxStartData         equ 0   ;parameter calculation at RxStartData
SPY_RxDataFFTidx        equ 0   ;FFT indexes for used tones
SPY_RxDataDiffPhases    equ 0   ;differential phases
SPY_RxDataSelPhase      equ -8  ;selected tone phase
SPY_RxDataSelDiffPhase  equ -8  ;selected tone differential phase
SPY_RxDataPhaseDev      equ 0   ;average phase deviations
SPY_RxDataScanMin       equ 0   ;results of the best tone pattern match
SPY_RxDataSoftSamples   equ 0
SPY_RxDataDeIntlvSamples equ 0
SPY_RxDataDecodeDev     equ 0
SPY_RxDataDecodeChar    equ 0

;Tx sensitive points:
SPY_TxCodecPtr  equ 0   ;the Tx CODEC pointer
SPY_TxStateRout equ 0   ;the Tx state routine
SPY_TxIFFT      equ 0   ;the result of the IFFT at the Tx
SPY_TxOutWindow equ 0   ;after windowing, before interpolation
SPY_TxAudio     equ 0   ;Tx audio output

SPY_RxDetune    equ  0.0 ;[Hz] Detune the Rx by that much at RxStartIdle
                         ;to see how is tunes back at the preamble
SPY_RxRandomDetune equ 0 ;randomly detune the Rx within the above range

SPY_TxRandSymbols equ 300 ;number of random symbols to be sent per transmition

;The following are monitoring options which put particular signals of state
;onto the external DSP pins of the B port. If you set the bit to
;a negative value you will effectively disable the option.
;Normally the bit assignments on the port B are:
;0 = PTT left
;1 = Tune UP left
;2 = Tune DOWN left
;3 = CAT out left
;4 = PTT right
;5 = Tune UP right
;6 = Tune DOWN right
;7 = CAT out right
;8-11 are inputs
;12 = ?
;13 = red status LED (all my modems use this for DCD signaling)
;14 = green status LED
;When you use the DSPCARD4/EVM56K with the left channel then the bits 4-7
;are not in use so you can take them for spying/debug.
;Beware of potential conflicts if you use bits 0-3 or 14.
;On the DSPCARD4 the bit 14 must be toggled every second or so
;to keep the watchdog happy.

MonRxAGC_up     equ  -4  ;monitor the gain changes being done by the AGC
MonRxAGC_down   equ  -6
MonRxCPU_bit    equ  -5  ;Monitor the CPU usage by the Rx code
MonTxCPU_bit    equ  -7  ;Monitor the CPU usage by the Tx code
MonRxSyncCatch_bit equ -4 ;The Rx "feels" the sync. tones

;This is the macro to facilitate the monitoring
Monitor macro mode,bit  ;clr/set/chg PB<bit> pin
        if bit>=0
          b\mode #bit,X:$FFE4
        endif
        endm

;*************************************************************************
;Control characters

NL      equ 0
CtrlA   equ 1
CtrlC   equ 3
CtrlD   equ 4
BELL    equ 7
LF      equ 10
CR      equ 13
CtrlQ   equ 17
CtrlR   equ 18
CtrlT   equ 20
CtrlU   equ 21

ReceiveSwitch equ CtrlR  ;to stop the transmitters
TransmitSwitch equ CtrlT ;to start the transmitter
TxAbortSwitch  equ CtrlA ;to abort a transmition
RxAbortSwitch  equ CtrlD ;to abort an Rx lock (force a resync.)
RxQualitySwitch equ CtrlQ ;to print out a signal quality report

;*************************************************************************
;Macros for pushing PTT, steering the LEDs.

      if RightChannel
PTT     macro mode      ;PTT line: clr/set/chg
        b\mode #4,X:$FFE4
        endm
      else
PTT     macro mode      ;PTT line: clr/set/chg
        b\mode #0,X:$FFE4
        endm
      endif

      if RightChannel
PushUp  macro mode      ;Push UP line of the TRX
        b\mode #5,X:$FFE4
        endm
      else
PushUp  macro mode
        b\mode #1,X:$FFE4
        endm
      endif

      if RightChannel
PushDown macro mode      ;Push DOWN line of the TRX
        b\mode #6,X:$FFE4
        endm
      else
PushDown macro mode
        b\mode #2,X:$FFE4
        endm
      endif

UpLED   macro mode      ;UP line (a red LED connected)
        b\mode #1,X:$FFE4
        endm

DownLED macro mode      ;DOWN line (a red LED connected)
        b\mode #2,X:$FFE4
        endm

YellowLED macro mode    ;CAT line (a yellow LED connected)
        b\mode #3,X:$FFE4
        endm

RedLED  macro mode      ; Red LED clr/set/chg
        b\mode #13,X:$FFE4
        endm


;macro for low-pass filtering on long variables
LongDecayAver macro Input,Aver,Scale
              sub Aver,Input
              rep Scale
                asr Input
              add Input,Aver
              endm

;*************************************************************************
;The actuall code (internal/external program RAM)

        LOMEM P:$0000
        HIMEM P:$1FFF

        org     p:user_code

        jmp <Initialize

DoFFT   ;routine to perform a complex FFT (and Inverse FFT too)
        ;copied from Motorola's Dr. BuB.
        ;alters a,b,x,y, rnm0, rnm1, n2, rnm4, rnm5, rnm6
        ;Uses 6 locations on System Stack

         move    #WindowLen/2,n0   ;initialize butterflies per group
         move    #1,n2             ;initialize groups per pass
         move    #WindowLen/4,n6   ;initialize C pointer offset
         move    #-1,m0            ;initialize A and B address modifiers
         move    m0,m1             ;for linear addressing
         move    m0,m4
         move    m0,m5
         move    #0,m6             ;initialize C address modifier for
                                   ;reverse carry (bit-reversed) addressing
;
; Perform all FFT passes with triple nested DO loop
;
         do      #WindowLenLog,_end_pass
         move    #FFTbuff,r0     ;initialize A input pointer
         move    r0,r4           ;initialize A output pointer
         lua     (r0)+n0,r1      ;initialize B input pointer
         move    #FFTcoef,r6     ;initialize C input pointer
         lua     (r1)-,r5        ;initialize B output pointer
         move    n0,n1           ;initialize pointer offsets
         move    n0,n4
         move    n0,n5

         do      n2,_end_grp
         move    x:(r1),x1  y:(r6),y0        ;lookup -sine and 
                                             ; -cosine values
         move    x:(r5),a   y:(r0),b         ;preload data
         move    x:(r6)+n6,x0                ;update C pointer

         do      n0,_end_bfy
         mac     x1,y0,b    y:(r1)+,y1       ;Radix 2 DIT
                                             ;butterfly kernel
         macr    -x0,y1,b   a,x:(r5)+    y:(r0),a
         subl    b,a        x:(r0),b     b,y:(r4)
         mac     -x1,x0,b   x:(r0)+,a  a,y:(r5)
         macr    -y1,y0,b   x:(r1),x1
         subl    b,a        b,x:(r4)+  y:(r0),b
_end_bfy
         move    a,x:(r5)+n5    y:(r1)+n1,y1   ;update A and B pointers
         move    x:(r0)+n0,x1   y:(r4)+n4,y1
_end_grp
         move    n0,b1
         lsr     b   n2,a1     ;divide butterflies per group by two
         lsl     a   b1,n0     ;multiply groups per pass by two
         move    a1,n2
_end_pass
        rts

;------------------------------------------------------------------------------
;Fast Hadamard transform (and the inverse)

;Attention: there is no guarantie that these two procedures perform
;the proper transform. The signs or the order of the data can be possibly
;inverted. I only care that the transformations are linear, reversible
;and orthogonal because this is what I need for encoding the data.

IFHT                    ;Y:(r1)/m1 points to the data to be encoded
        move #>1,x1     ;m1 = number of samples - 1 (must be 2^N-1)
        move m1,a       ;I intend to use this function for encoding the data
        add x1,a        ;thus I normally place all zeros but one single value
IFHT_pass               ;in the input array and after IFHT I get the vector to
        lsr a1          ;to be sent out
        move a1,n1
        jcs <IFHT_ret           ;Beware that the data is being summed up and
        .loop x1                ;the numbers can grow as high as 2^N times.
          .loop n1              ;When arythmetic saturation comes into action
            move Y:(r1),x0      ;the result can be one big mess. Thus you
            tfr x0,b Y:(r1+n1),a ;must make sure the data is scaled right.
            add x0,a a,x0
            sub x0,b a,Y:(r1+n1)
            move b,Y:(r1)+
          .endl
          move (r1)+n1
        .endl
        tfr x1,b
        asl b n1,a
        move b,x1
        jmp <IFHT_pass
IFHT_ret                ;x0,x1,a,b,n1 are modified
        rts             ;Y:(r1)/m1 contains the transformed samples
                        ;r1 should not change

FHT                     ;Y:(r1)/m1 points to the data to be decoded
        move #>1,x1     ;m1 like for the IFHT must be 2^N-1
        move m1,a
        add x1,a #<1,n1
FHT_pass
        lsr a1
        move a1,x1
        jcs <IFHT_ret
        .loop x1
          .loop n1
            move Y:(r1),x0
            tfr x0,b Y:(r1+n1),a
            sub x0,a a,x0
            add x0,b a,Y:(r1+n1)
            move b,Y:(r1)+
          .endl
          move (r1)+n1
        .endl
        move n1,b
        asl b x1,a
        move b,n1
        jmp <FHT_pass

;------------------------------------------------------------------------------

ProcessLoop_wait
        wait
ProcessLoop     ;main loop to process CODEC's samples
      Monitor clr,MonTxCPU_bit
      Monitor clr,MonRxCPU_bit
        move X:<RxSliceSeparDeci,x0     ;here we calculate:
        move #>DecimateRatio,y0         ;DecimateRatio*RxSliceSeparDeci
        mpy x0,y0,b X:<RxCodecPtr,r2
        asr b
        move b0,a
        move a,X:<RxSliceSeparSamp
        jsr <CheckSampleBlock   ;look for that many samples to come
        move #CodecBuffLen*4-1,m2 ;set m2 so r2 wraps correctly
        jcc <RxProcess          ;if there are enough samples we do
                                ;the Rx processing
        move #>DecimateRatio*TxSliceSepar,a ;similar stuff for Tx
        move X:<TxCodecPtr,r2   ;but the SliceSepar is fixed for the Tx
        jsr <CheckSampleBlock   ;
        jcc <TxProcess          ;do Tx processing if enough samples
        jmp <ProcessLoop_wait   ;wait if not yet...

;------------------------------------------------------------------------------

RxProcess       ;r2/m2 points to the samples we should process
      Monitor set,MonRxCPU_bit

      if SPY&&SPY_RxCodecPtr
        move r2,a
        move #>CodecBuff,x0
        sub x0,a
        rep #8
          asl a
        jsr <SpyA
      endif
      if SPY&&SPY_RxCarFreq
        move X:<RxCarFreq,a
        rep #8
          asl a
        jsr <SpyA
      endif
      if SPY&&SPY_RxSliceSepar
        move X:<RxSliceSeparDeci,a
        move #>RxSliceSepar,x0
        sub x0,a
        rep #8
          asl a
        jsr <SpyA
      endif

        move #<4,n2                ;for easier input addressing
        move #<AliasFilterLen-1,m4
      if RightChannel
        move (r2)+
      endif
      if SPY&&SPY_RxAudio
       move r2,r0
       move #<4,n2
       .loop X:<RxSliceSeparSamp
         move X:(r2)+n2,a
         jsr <SpyA
         nop
       .endl
       move r0,r2
      endif
      if RxAGC
       if RxAGCavoidPTT
        PTT tst         ;avoid any AGC level measurement or gain change
        jcs <AGC_MeasureDone ;while PTT is on.
       endif
        move X:<RxAudioPeak,a   ;update the peak value of the input signal
        clr b r2,r0             ;save r2, clear power sum
        .loop X:<RxSliceSeparSamp ;look through the new samples
          move X:(r2)+n2,x0     ;and find the largest one
          cmpm x0,a
          tlt x0,a
          mac x0,x0,b           ;sum up the power by the way
        .endl
        abs a r0,r2             ;restore r2
        move a,X:<RxAudioPeak
        rep #WindowLenLog-2
          asr b
        rnd b #>(WindowLen/4.0)/(DecimateRatio*RxSliceSepar),x1
        move b,x0
        mpyr x0,x1,b
        rnd b X:<RxAudioMS,x1
        move b,x0 #RxAGC_MSfollow,y0
        mpy x0,y0,b #1.0-RxAGC_MSfollow,y0
        macr x1,y0,b
        move b,X:<RxAudioMS
AGC_MeasureDone
      endif ;of RxAGC
        move X:<WindowInpTapPtr,r1 ;r1 to address the input window tap
        move #WindowLen-1,m1
        move #<AliasFilterLen-1,m4
        .loop X:<RxSliceSeparDeci ;filter, decimate and fill the window's tap
          move #4*(AliasFilterLen-DecimateRatio),n2
          move #AliasFilterInpI,r4 ;r4 to address the input filter shape
          move (r2)-n2            ;move back in the CODEC input buffer
          move #<4,n2             ;n2=4 for easier input adressing
          move r2,x1              ;save r2 (where we start the filter tap)
          clr a X:(r2)+n2,x0 Y:(r4)+,y0
          .loop #AliasFilterLen-1
            mac x0,y0,a X:(r2)+n2,x0  Y:(r4)+,y0
          .endl
          macr x0,y0,a #AliasFilterInpQ,r4
          move x1,r2      ;restore the tap start
          nop
          clr b X:(r2)+n2,x0 Y:(r4)+,y0
          .loop #AliasFilterLen-1
            mac -x0,y0,b X:(r2)+n2,x0  Y:(r4)+,y0
          .endl
          macr -x0,y0,b a,X:(r1) ;a,b = decimated sample (I/Q), restore r2
          move b,Y:(r1)          ;put new sample into the tap
          move X:<RxCarPhase,x0  ;get the receiver NCO's carrier phase
          move X:<RxCarFreq,a    ;advance the phase
	  add x0,a
	  move a1,X:<RxCarPhase
	  jsr <IQ               ;compute I and Q (modifies a,b,x,y,r0,m0,n0)
          move b,y0
          move X:(r1),x1 a,y1
          mpy x1,y1,a Y:(r1),x0 ;mix the input samples with the NCO
          macr -x0,y0,a
          mpy x1,y0,b
          macr x0,y1,b a,X:(r1)
          move b,Y:(r1)+
        if SPY&&SPY_RxIQ
          jsr <SpySync
          .if <cc>
            jsr <SpyA
            tfr b,a
            jsr <SpyA
          .endi
          nop
        endif
        .endl
      if RightChannel
        move (r2)-
      endif
        move r2,X:<RxCodecPtr     ;save the CODEC buffer pointer

RxSlide         ;apply window, copy data to FFTbuff
        move #DataWindowInp,r5
        move #WindowLen-1,m5
        move #FFTbuff,r0
        move #WindowLen-1,m0
        .loop #WindowLen                ;apply input window
          move X:(r1),x0 Y:(r5)+,y0     ;x0 = sample_I, y0 = input window
          mpyr x0,y0,a Y:(r1)+,x0       ;a = x0*y0, x0 = sample_Q
          mpyr x0,y0,a a,X:(r0)         ;a = x0*y0, save windowed I
          move a,Y:(r0)+                ;save windowed Q into FFTbuff
        .endl
        move r1,X:<WindowInpTapPtr

        jsr <DoFFT              ;execute FFT

      if RxAGC                  ;update the signals power (mean square)
       if RxAGCavoidPTT
        PTT tst         ;avoid any AGC level measurement or gain change
        jcs <CheckMS_OK ;while PTT is on.
       endif

       if SPY&&SPY_PeakMS
        clr b X:<RxAudioPeak,a
      if !RxAGCsensePeak
        move b,X:<RxAudioPeak
      endif
        jsr <SpyA
        move X:<RxAudioMS,a
        jsr <SpyA
       endif

      Monitor clr,MonRxAGC_down
      Monitor clr,MonRxAGC_up

      if RxAGCavoidDCD
       jset #RxSta_DCD,X:<RxState,CheckMS_OK
      endif

      move X:<RxCodecPtr,r2     ;set r2 to address the CODEC's buffer
      move #CodecBuffLen*4-1,m2 ;set m2 so r2 wraps in the CODEC's buffer
                                ;we will need r2/m2 for RxGainUp/Down
      if RxAGCsensePeak
        move #RxAGCmaxPeak,x0   ;check if peak is above the given max.
        move X:<RxAudioPeak,b
        cmp x0,b #<0,b
        move b,X:<RxAudioPeak   ;clear the peak
        jcc <GainDown           ;reduce gain if the peak is too high
      endif

        move X:<RxAudioMS,a
        move #RxAGCminMS,x0     ;see, if the MS is below the minimal level
        cmp x0,a
        jcc <CheckMSmax         ;if not, then check if above the max. level
                                ;if so, increase the CODEC's input gain
        move X:<RxAGCcount,a    ;decrement the timeout
        move #>1,x0
        sub x0,a 
        move a,X:<RxAGCcount
        jgt <CheckMS_OK         ;leave if not yet zero
        jsr <RxGainUp           ;increase the CODEC's gain
        jcs <CheckMS_OK
        move #0.7071,y0         ;increase the MS to follow
        move X:<RxAudioMS,x0    ;the gain change faster
        mpyr x0,y0,a #>RxAGCfall,x0 ;preset the AGC "timer" to prohibit
        asl a x0,X:<RxAGCcount      ;gain increase for the RxAGCfall period
        move a,X:<RxAudioMS
        Monitor set,MonRxAGC_up
        jmp <CheckMS_OK

CheckMSmax                      ;is the MS above the given maximum
        move #>RxAGChold,y0     ;preset the AGC "timer" to prohibit gain increase
        move #RxAGCmaxMS,x0     ;for the RxAGChold period
        cmp x0,a y0,X:<RxAGCcount ;MS above maximum ?
        jcs <CheckMS_OK         ;if not then we are done
GainDown
        jsr <RxGainDown         ;otherwise decrease the CODEC's gain
        jcs <CheckMS_OK
        move #0.7071,y0         ;decrease the MS to follow expected
        move X:<RxAudioMS,x0    ;gain reduction faster
        mpyr x0,y0,a #>RxAGChold,x0     ;initialize the AGC hold count-down
        move a,X:<RxAudioMS
        move x0,X:<RxAGCcount
        Monitor set,MonRxAGC_down

CheckMS_OK

      endif     ;of RxAGC

        if SPY&&SPY_RxFFT            ;SPY on the FFT result on the Rx signal
         jsr <SpySync
         .if <cc>
          move #FFTbuff,r0
          move #<0,m0
          move #WindowLen/2,n0
          .loop #WindowLen
            move X:(r0),a
            jsr <SpyA
            move Y:(r0)+n0,a
            jsr <SpyA
            nop
          .endl
         .endi
        endif
        if SPY&&SPY_RxFFTpower
         jsr <SpySync
         .if <cc>
          move #FFTbuff,r0
          move #<0,m0
          move #WindowLen/2,n0
          .loop #WindowLen
            move X:(r0),x0
            mpy x0,x0,a Y:(r0)+n0,x0
            mac x0,x0,a
            jsr <SpyA
            nop
          .endl
         .endi
        endif

        move #WindowLen-1,m2
        move  X:<RxFFTphaseCorr,r2
        move  X:<RxSliceSeparDeci,n2
        move #FFTbuff,r0
        move  (r2)+n2
        move  r2,X:<RxFFTphaseCorr
        move  r2,n2
        move #<0,m0
        move #WindowLen/2,n0
        move #SinTable,r2               ;r2/m2/n2 for correcting FFT phases
        .loop #WindowLen                ;correct the phases of the FFT
          move X:(r0),x1
          move X:(r2),y1
          mpy x1,y1,a Y:(r2)-n2,y0
          mpy x1,y0,b Y:(r0),x0
          macr -x0,y0,a
          macr x0,y1,b a,X:(r0)
          move b,Y:(r0)+n0
        .endl

      if SPY&&SPY_RxCorrFFT
        jsr <SpySync
        .if <cc>
          .loop #WindowLen
            move X:(r0),a
            jsr <SpyA
            move Y:(r0)+n0,a
            jsr <SpyA
            nop
          .endl
          .loop #WindowLen/2
            move (r0)+n0
          .endl
        .endi
      endif

      if SPY&&(SPY_RxSelFFTbin>=0)
        jsr <SpySync
        .if <cc>
          move #@rvb(SPY_RxSelFFTbin,WindowLenLog),n0
          nop
          move X:(r0+n0),a
          jsr <SpyA
          move Y:(r0+n0),a
          jsr <SpyA
          move #WindowLen/2,n0
        .endi
      endif
      if SPY&&(SPY_RxSelFFTphase>=0)
        move #@rvb(SPY_RxSelFFTphase,WindowLenLog),n0
        nop
        move X:(r0+n0),a
        move Y:(r0+n0),b
        jsr <RaughNormIQ
        jsr <Phase
        jsr <SpyA
        move #WindowLen/2,n0
      endif

;here spectral data is available in the FFTbuff (bit-reversed) at r0/m0
;for the Rx processing
        move X:<RxStateRoutine,r2
        nop
        jmp (r2)

;        jmp <ProcessLoop        ;end of Rx processing - return to the main loop

;------------------------------------------------------------------------------

TxProcess
    if !SPY
      jsr <CheckInput   ;check if any characters from the user, if so then
    endif               ;interprete them or store them in the TxCharBuff

      Monitor set,MonTxCPU_bit
      if SPY&&SPY_TxCodecPtr
        move r2,a
        move #>CodecBuff,x0
        sub x0,a
        rep #8
          asl a
        jsr <SpyA
      endif

;here the Tx should prepare data in FFTbuff and jump to TxDoFFT
;or jump straigth away to TxDoSilence

        if SPY&&SPY_TxStateRout
          move X:<TxStateRoutine,a
          rep #8
            asl a
          jsr SpyA
        endif
        move X:<TxStateRoutine,r0
        bchg #TxSta_SkipSlice,X:<TxState
        jcs <TxDoSilence
        jmp (r0)

TxDoSilence
        move X:<WindowOutTapPtr,r1 ;r1 to address the output window tap
        move #WindowLen-1,m1
        move #WindowLen-TxSliceSepar,n1
        nop
        clr a (r1)+n1
        .loop #TxSliceSepar
          move a,L:(r1)+
        .endl
        jmp <TxDoOutput

TxDoFFT  jsr <DoFFT             ;execute IFFT

        if SPY&&SPY_TxIFFT      ;SPY on the IFFT results on the Tx signal
         jsr <SpySync
         .if <cc>
          move #FFTbuff,r0
          move #<0,m0
          move #WindowLen/2,n0
          .loop #WindowLen
            move X:(r0),a
            jsr <SpyA
            move Y:(r0)+n0,a
            jsr <SpyA
            nop
          .endl
         .endi
        endif

TxDoWindow                      ;overlap the new window
        move #FFTbuff,r0        ;(r0)/m0/n0 = IFFT result
        move #<0,m0             ;bit-reversly addressed
        move #WindowLen/2,n0
        move #DataWindowOut,r5   ;(r5)/m5 = Tx output window (shape)
        move #WindowLen-1,m5
        move X:<WindowOutTapPtr,r1 ;r1 to address the output window tap
        move #WindowLen-1,m1
        move X:(r0),x0 Y:(r5)+,y0 ;x0=first data(re), y0=first window
        .loop #WindowLen-TxSliceSepar   ;take the IFFT result and overlay it
          move L:(r1),ab                ;over the TxWindow tap
          macr x0,y0,a Y:(r0)+n0,x0     ;but before shape it by TxWindow
          macr x0,y0,b X:(r0),x0 Y:(r5)+,y0
          move ab,L:(r1)+
        .endl
        .loop #TxSliceSepar-1
          mpyr x0,y0,a Y:(r0)+n0,x0
          mpyr x0,y0,b X:(r0),x0 Y:(r5)+,y0
          move ab,L:(r1)+
        .endl
        mpyr x0,y0,a Y:(r0)+n0,x0
        mpyr x0,y0,b a,X:(r1)
        move b,Y:(r1)+
TxDoOutput                              ;(r1)/m1 = Tx window tap
        move X:<TxCodecPtr,r2           ;(r2)/m2 = CODEC pointer
        move #CodecBuffLen*4-1,m2
        move #AliasFilterOutI,r4        ;(r4)/m4/n4 = anti-alias filter
        move #<AliasFilterLen-1,m4
        move #<AliasFilterOutQ-AliasFilterOutI,n4
        move (r2)+                      
      if RightChannel
        move (r2)+
      endif
        .loop #TxSliceSepar    ;output interpolation
        if SPY&&SPY_TxOutWindow
         jsr <SpySync
         .if <cc>
          move X:(r1),a
          jsr SpyA
          move Y:(r1),a
          jsr SpyA
         .endi
        endif
          move #4*(AliasFilterLen-DecimateRatio),n2
          move L:(r1)+,y        ;y1,y0 = processed samples (I/Q) ready to be output
          move (r2)-n2          ;move back in the CODEC input buffer
          move #<4,n2           ;n2=4 for easier output adressing
        if SPY&&SPY_TxAudio
          nop
          move (r2)-n2
          move Y:(r2)+n2,a
          jsr <SpyA
        endif
          .loop #AliasFilterLen-DecimateRatio
            move X:(r4),x0 Y:(r2),a
            mac x0,y1,a X:(r4+n4),x0
            macr x0,y0,a (r4)+
            move a,Y:(r2)+n2      ;output to the left channel
          .endl
          .loop #DecimateRatio
            move X:(r4),x0
            mpy x0,y1,a X:(r4+n4),x0
            macr x0,y0,a (r4)+
            move a,Y:(r2)+n2
          .endl
          nop
        .endl
        move (r2)-
      if RightChannel
        move (r2)-
      endif
        move r2,X:<TxCodecPtr
        move r1,X:<WindowOutTapPtr

        jmp <ProcessLoop

;==============================================================================
;State machine routines for the receiver

RxCarrierAbort
       if (!SPY)&&RxReport
         move #$FFFF,m1
         move #RxLockAbortMsg,r1
         jsr <PutString
       endif
        bclr #RxSta_AbortReq,X:<RxState ;clear the abort request

RxStartIdle                             ;carrier lost or Rx (re)initialized
        bclr #RxSta_DCD,X:<RxState      ;declare DCD OFF
        RedLED clr                      ;turn off the red LED
        Monitor clr,MonRxSyncCatch_bit

        move #>RxDoIdle,x0      ;set Rx handler to "idle"
        clr a x0,X:<RxStateRoutine
;the following is to mistune the Rx to see if it corrects properly
      if SPY&&(SPY_RxDetune!=0.0)
       if SPY_RxRandomDetune
        jsr <Rand48
        move a0,x0
        eor x0,a1 #SPY_RxDetune*DecimateRatio*2.0/SampleFreq,y0
        move a1,x0
        mpyr x0,y0,a
       else
        move #SPY_RxDetune*DecimateRatio*2.0/SampleFreq,a
       endif
      endif
        move a,X:<RxCarFreq     ;zero the tuning corrector
        move #-1.0,a            ;set the acceptance counter to -1.0
        move a,X:<RxAcceptCount ;
        move #<$7F,b
        clr a #RxSyncPhaseDev,r0        ;set the phase deviation counters
        move #4*RxSyncPhaseDevLen-1,m0  ;deviation to +0.99
        move #RxSyncPhaseDevMid-RxSyncPhaseDev,n0
        .loop #4*RxSyncPhaseDevLen      ;and shift to 0
          move ab,L:(r0+n0)
          move ab,L:(r0)+
        .endl
        move r0,X:<RxPhaseDevPtr
        clr a #RxSyncPhasePipe,r0            ;clear the FFT phase pipe
        move #4*RxSyncScanLen-1,m0
        .loop #4*RxSyncScanLen
          move a,Y:(r0)+
        .endl
        move r0,X:<RxPipePtr            ;and reset the pipe pointer
        move a,X:<RxFFTphaseCorr        ;Reset the FFT phase corrector
        move #>RxSliceSepar,x0
        move x0,X:<RxSliceSeparDeci     ;set the default FTT slice separation
        jmp <ProcessLoop

RxDoIdle        ;keep correlations up to date and hunt for the sync. tones
                ;r0/n0/m0 points to the FFT data

        jset #RxSta_AbortReq,X:<RxState,RxCarrierAbort ;if the user wishes to abort
      if RxSyncAvoidPTT
        PTT tst         ;force continues restart if the PTT is on - it is
        jcs <RxStartIdle ;a brutal way to avoid capturing his own transmition
      endif
        move #>RxSliceSepar,x0
        move x0,X:<RxSliceSeparDeci     ;set the default FTT slice separation
        move #RxSyncPhaseBuff,r4        ;r4/m4 points to phase buffer
        move #RxSyncScanLen-1,m4
        move X:<RxPipePtr,r5            ;r5/m5 points to the FFT phase pipe
        move #@rvb(RxSyncScanFirst,WindowLenLog),n0
        move #4*RxSyncScanLen-1,m5
        move (r0)+n0                    ;advance to the first FFT bin we care about
        move #WindowLen/2,n0
        .loop #RxSyncScanLen
          move L:(r0)+n0,ab             ;take the FFT I/Q
          jsr <RaughNormIQ              ;and compute the phase
          jsr <Phase
          move Y:(r5),x0                ;read the phase 4 slices before this one
          sub x0,a a,Y:(r5)+            ;and compute the difference
          move a1,Y:(r4)+               ;save the phase difference
        .endl
        move r5,X:<RxPipePtr
      if SPY&&(SPY_RxSyncSelDiffPhase>=0)
         move #SPY_RxSyncSelPhase,n4
         move Y:(r4+n4),a
        jsr <SpyA
      endif
      if SPY&&SPY_RxSyncPhases
        jsr <SpySync
        .if <cc>
          .loop #RxSyncScanLen
            move Y:(r4)+,a
            jsr <SpyA
            nop
          .endl
        .endi
      endif
        move X:<RxPhaseDevPtr,r5
        move #4*RxSyncPhaseDevLen-1,m5
        move #RxSyncPhaseDevMid-RxSyncPhaseDev,n5
        move #FFTbuff,r0                ;here we re-use the FFTbuff
        move #DataCarriers/2-1,m0       ;because we don't need the FFT data
        move #<2*DataCarrSepar,n4       ;there anymore
        .loop #RxSyncPhaseDevLen        ;for posible carrier positions
          move r4,n0                    ;we select the FFT bins
          .loop #DataCarriers/2
            move Y:(r4)+n4,a
            asl a
            move a1,Y:(r0)+
          .endl
          move n0,r4
          jsr <AverMS ; or AverMS_rej   ;and we compute the average
                                        ;phase shift and deviation
          move a,x0 #>RxSyncAverWeight,y0       ;x0 = input
          mpy x0,y0,a #<1.0-RxSyncAverWeight,y1
          mac x0,y0,a X:(r5+n5),x0              ;x0 = middle elem.
          mac x0,y1,a X:(r5),x1                 ;x1 = output
          macr -x1,y0,a #<4*RxSyncAverWeight,y0 ;a = updated middle elem.
          mpy x0,y0,a a,X:(r5+n5)
          move #<1.0-4.0*RxSyncAverWeight,y1
          macr x1,y1,a                          ;a = updated output
          move a,X:(r5)
          move b,x0 #>RxSyncAverWeight,y0       ;x0 = input
          mpy x0,y0,b #<1.0-RxSyncAverWeight,y1
          mac x0,y0,b Y:(r5+n5),x0              ;x0 = middle elem.
          mac x0,y1,b Y:(r5),x1                 ;x1 = output
          macr -x1,y0,b #<4*RxSyncAverWeight,y0 ;a = updated middle elem.
          mpy x0,y0,b b,Y:(r5+n5)
          move #<1.0-4.0*RxSyncAverWeight,y1
          macr x1,y1,b                          ;a = updated output
          move b,Y:(r5)+
;          move a,x1 #>RxSyncAverWeight,y0 ;phase shift and deviation
;          mpy x1,y0,a X:(r5),x0
;          move #<1.0-RxSyncAverWeight,y1
;          macr x0,y1,a b,x1
;          mpy x1,y0,b Y:(r5),x0
;          macr x0,y1,b a,X:(r5)
;          move b,Y:(r5)+
          move (r4)+
        .endl
        move r5,X:<RxPhaseDevPtr
        if SPY&&SPY_RxSyncMinPhases
         jsr <SpySync
         .if <cc>
          move r4,n0
          move Y:-(r5),a
          move (r4)-
          move r4,r2
          .loop #RxSyncPhaseDevLen-1
            move Y:-(r5),x0
            cmp x0,a (r4)-
            tgt x0,a r4,r2
          .endl
          move r2,a
          move #RxSyncPhaseBuff,x0
          sub x0,a r2,r4
          rep #8
            asl a
          jsr <SpyA
          move #<2*DataCarrSepar,n4
          .loop #DataCarriers/2
            move Y:(r4)+n4,a
            asl a
            move a1,x0
            tfr x0,a x0,Y:(r0)+
            jsr <SpyA
            nop
          .endl
          jsr <AverMS
          jsr <SpyA
          tfr b,a
          jsr <SpyA
          move #<RxSyncPhaseDevLen,n5
          move n0,r4
          move (r5)+n5
         .endi
        endif
        if SPY&&SPY_RxSyncPhaseDev
          move r5,a
          move #>RxSyncPhaseDev,x0
          sub x0,a
          .if <eq>
            jsr <SpySync
            .if <cc>
              .loop #4*RxSyncPhaseDevLen
                move X:(r5),a
                jsr <SpyA
                move Y:(r5)+,a
                jsr <SpyA
                nop
              .endl
            .endi
          .else
            rep #8
              asl a
            jsr <SpyA
          .endi
        endif
        move #<RxSyncPhaseDevLen,n5             ;here we look for the
        move #<2*RxSyncScanMargin+1-1,m0        ;lowest phase deviation
        move #<0,r0     ;r0 will count FFT bins
        move (r5)-n5
        move #<4-1,m1
        move #<0,r1     ;r1 will count symbol positions
        move #<2*RxSyncPhaseDevLen+DataCarrSepar,n5
        move Y:(r5),a
        move Y:(r5+n5),x0
        add x0,a r0,n0  ;a = minimum level, n0 = FFT bin where minimum
        move r1,n1      ;n1 = time where minimum
        move r5,y1      ;y1 = where in memory
        .loop #4        ;scan through 4 symbol positions
          move #<2*RxSyncPhaseDevLen+DataCarrSepar,n5
          .loop #2*RxSyncScanMargin+1 ;scan through FFT bins
            move Y:(r5+n5),b
            move Y:(r5),x0
            add x0,b r5,x0
            cmp a,b (r5)+
            .if <lt>            ;if lower dev. found then
              tfr b,a r0,n0     ;register its value and position
              move r1,n1
              move x0,y1
            .endi
            move (r0)+
          .endl
          move #<DataCarrSepar,n5
          move (r1)+
          move (r5)+n5
        .endl           
        move y1,r5           ;a = phase Mean Square deviation at best point
        asr a #<2*RxSyncPhaseDevLen+DataCarrSepar,n5
        move X:(r5),b
        move X:(r5+n5),x0
        add x0,b a,X:<RxSignalMinDev
        asr b #RxSyncCatchThres,x1  ;b = phase offset at lowest deviation
        cmp x1,a b,X:<RxTmpI        ;we will use it to correct the frequency offset
        move X:<RxAcceptCount,a
        move #<(3*RxSyncAverWeight)/8,x0
        .if <lt>
          add x0,a
        .else
          sub x0,a
        .endi
        move a,X:<RxAcceptCount
      if SPY&&SPY_RxSyncScanMin
        move X:<RxSignalMinDev,a ;lowest MS deviation
        jsr <SpyA
        move X:<RxTmpI,a        ;measured frequency offset
        jsr <SpyA
        move X:<RxAcceptCount,a
        jsr <SpyA
        move X:<RxCarFreq,a     ;current (known) software freq. offset
        jsr <SpyA
        move n0,a               ;carrier +/-
        move #>RxSyncScanMargin,x0
        sub x0,a
        rep #8
          asl a
        jsr <SpyA
        move n1,a               ;measured time offset
        rep #8
          asl a
        jsr <SpyA
      endif
        move n1,b       ;check if we are close to the optimal sampling ?
        tst b X:<RxAcceptCount,a ;n1=0 in such case.
        jne <ProcessLoop
        move #<$7F,x1
        cmp x1,a        ;of accept counter not high enough
        jlt <ProcessLoop ;then go on with the current phase
        move n0,a
        tst a #>2*RxSyncScanMargin+1,x0
        jeq <ProcessLoop ;do NOT lock when the frequency is at the edge
        cmp x0,a         ;of the scanned area because in such case
        jeq <ProcessLoop ;we are could capture only a part of the carriers
        move n5,n4       ;otherwise compute approx. timing shift
        move Y:(r5+n5),x0 ;Y:(r5)/m5/n5 still points to phase MS deviations
        move #<RxSyncPhaseDevLen,n5
        move m5,m4
        lua (r5)-n5,r4
        move Y:(r5),b
        add x0,b Y:(r4+n4),x0 ;b = lowest deviation -> x1
        move Y:(r4),a         ;
        lua (r5)+n5,r4
        add x0,a b,x1         ;a = deviation one slice before -> y0
        move Y:(r4+n4),x0
        move Y:(r4),b
        add x0,b a,y0         ;b = deviation one slice after -> y1
        sub a,b  b,y1         ;b = the difference
        move #>RxSliceSepar/4,a
        .if <gt>
          neg a
        .endi
        move #>RxSliceSepar,b
        add b,a
        move a,X:<RxSliceSeparDeci

      if (!SPY)&&RxReport
        move #$FFFF,m1
        move n0,a                       ;a = carrier +/-
;        move #>RxSyncScanMargin+8,a ;*** DEBUG ***
        move #>RxSyncScanMargin,x0
        sub x0,a X:<RxTmpI,x0           ;x0 = frequency offset measured
        move a,x1 #>2.0/WindowLen/2,y0
        mpy x1,y0,a #>1.0/SymbolSepar/2,y0 ;
        move a0,a
;        move #<0,x0 ;*** DEBUG ***
        macr x0,y0,a #RxLockMsg_freq,r1   ;the sign here is not obvious...
        move a,x0
        jsr <FormatFreq_sdd
        move X:<RxSignalMinDev,a
;        asl a a,x0
;        add x0,a
;        move #>0.001,a ;*** DEBUG ***
        jsr <Log2 ;modifies r0,x,b
        neg a #RxLockMsg_SNR,r1
        rep #6
          asr a
        move a,x0
        jsr <FormatPosLog2frac_ddd_dB
        move #RxLockMsg,r1
        jsr <PutString
      endif

      jmp <RxStartData

;The following computes the average and mean square of DataCarriers/2 numbers
AverMS                          ;Y:(r0)/m0 = data to be analyzed
        clr b Y:(r0)+,x0
        .loop #DataCarriers/2-1
          add x0,b Y:(r0)+,x0
        .endl
        add x0,b
        rep #DataCarriersLog-1-1
          asr b                 ;b = the simple average
        asr b Y:(r0)+,a
        clr b b,x1              ;move average to x1
        .loop #DataCarriers/2
          sub x1,a              ;subtract the average
          move a,x0
          mac x0,x0,b Y:(r0)+,a ;sum up the difference squares
        .endl
        rep #DataCarriersLog-1
          asr b
        tfr x1,a (r0)-
        rts                     ;a = average, b = mean square
                                ;modifies x0,x1
                                ;r0/m0 should not change
      if 0
AverMS_rej                      ;Y:(r0)/m0 - data to be analyzed
        clr b Y:(r0)+,x0        ;this routine is more smart - it attempts
        .loop #DataCarriers/2-1 ;to reject the numbers which "stick out".
          add x0,b Y:(r0)+,x0
        .endl
        add x0,b
        rep #DataCarriersLog-1-1
          asr b                 ;b = the simple average
        asr b Y:(r0),a
        clr b b,y0              ;move average to y0
        .loop #DataCarriers/2
          sub y0,a              ;subtract the average
          move a,x0 a,Y:(r0)+   ;save back the corrected data
          mac x0,x0,b Y:(r0),a  ;sum up the difference squares
        .endl
        rep #DataCarriersLog-1-3
          asr b                 ;b = MS * 8, y0 = average
      .loop #3          ;we can try several iterations...
        clr a b,y1
        .loop #DataCarriers/2
          tfr y1,b Y:(r0),x0
          mac -x0,x0,b
          .if <mi>              ;if x0*x0 > b
             tfr x0,b           ;then cut this number by half
             asr b              ;(fastest thing we can do...)
             move b,x0 b,Y:(r0)
          .endi
          add x0,a (r0)+        ;sum up the numbers
        .endl
        rep #DataCarriersLog-1
          asr a                 ;a = 1st average correction
        add y0,a a,x1           ;add the previous average
        clr b a,y0               ;y0 = new average, x1 = correction
        .loop #DataCarriers/2
          sub x1,a              ;subtract the correction
          move a,x0 a,Y:(r0)+   ;save back the corrected data
          mac x0,x0,b Y:(r0),a  ;sum up the difference squares
        .endl
        rep #DataCarriersLog-1-3
          asr b
      .endl
        asr b
        asr b
        asr b y0,a
        rts                     ;a = average, b = mean square
                                ;modifies x0,x1 and data too (!)
                                ;r0/m0 should not change
      endif ; of 0

;------------------------------------------------------------------------------

;OK. here we finally decided to go into the data phase, what we know is:
;n0 = FFT bin where the first carrier is about
;X:<RxTmpI = precise frequency offset:
; offset +0.5 Hz (FTT bin is 3.125 Hz) makes +13030/2/32768 = SymbolSepar/(WindowLen/4)*0.5/3.125
;we are at the closest time to the proper symbol sampling
;RxSyncPhasePipe contains the older phases pointed by X:<RxPipePtr
;the data starts at RxSyncScanFirst FFT bin.

RxStartData
        bset #RxSta_DCD,X:<RxState      ;declare DCD ON
        RedLED set                      ;turn on the red LED

        move X:<RxTmpI,x0               ;x0 = frequency offset measured
        move #>1.0/SymbolSepar/2,x1     ;during RxDoIdle
        move X:<RxCarFreq,a             ;
        macr -x0,x1,a                   ;we move the RxCarFreq to compensate
        move a,X:<RxCarFreq             ;for this offset

        move #>RxDoData,x0              ;set the state routine to RxDoData
        clr a x0,X:<RxStateRoutine
        move a,X:<RxSymbolPhase         ;we enter RxStartData at about
                                        ;the symbol sampling time
        clr a #RxDataPhasePipe,r5       ;clear the FFT phase pipe
        move #4*RxDataScanLen-1,m5
        .loop #4*RxDataScanLen
          move a,Y:(r5)+
        .endl
        move r5,X:<RxPipePtr            ;set the pipe pointer

        clr a #RxDataPhaseDev,r5        ;set the phase deviation counters
        move #<4*RxDataPhaseDevLen-1,m5 ;deviation to 0
        move #<RxDataPhaseDevMid-RxDataPhaseDev,n5
        .loop #4*RxDataPhaseDevLen      ;and shift to 0
          move a,L:(r5+n5)
          move a,L:(r5)+
        .endl
        move r5,X:<RxPhaseDevPtr

        clr a #RxDecodeDev,r5
        move #<2*2*RxDataScanMargin-1,m5
        move #<RxDecodeDevMid-RxDecodeDev,n5
        .loop #2*2*RxDataScanMargin
          move a,L:(r5+n5)
          move a,L:(r5)+
        .endl
        move r5,X:<RxDecodeDevPtr

        move #>RxSyncScanFirst,a
        move n0,x0              ;n0 = carrier offset -> x0
        add x0,a #>2*RxDataScanMargin*DataCarrSepar,x0
        sub x0,a #<0,b          ;a = the first bin of the FFT we care about
        move a,X:<RxDecodeFirstBin
        .loop #WindowLenLog     ;for the data phase
          lsr a1                ;reverse the bits (FFT results
          rol b1                ;are addressed with reverse carry)
        .endl
        move b1,X:<RxDataFirstBin

        clr a #RxDataSamplePipe,r2      ;clean the data pipe
        move #DataInterleave*2*RxDataSampleLen-1,m2
        .loop #DataInterleave*2*RxDataSampleLen
          move a,Y:(r2)+
        .endl
        move r2,X:<RxDataSamplePtr      ;and setup the pointer

      if SPY&&SPY_RxStartData
        move n0,a
        rep #8
          asl a
        jsr <SpyA
        move r1,a
        rep #8
          asl a
        jsr <SpyA
        move #>@rvb(FirstDataCarr-2*RxDataScanMargin*DataCarrSepar,WindowLenLog),a
        rep #8
          asl a
        jsr <SpyA
        move X:<RxCarFreq,a
        jsr <SpyA
      endif
        move #<$7F,a
        move a,X:<RxAcceptCount
        move a,X:<RxTimeCorrHold

        jmp <ProcessLoop

RxDoData                                ;L:(r0)/m0 = FFT results
        jset #RxSta_AbortReq,X:<RxState,RxCarrierAbort ;if the user wishes to abort
        move #>RxSliceSepar,x0
        move x0,X:<RxSliceSeparDeci     ;set the default FTT slice separation
        move X:<RxSymbolPhase,r5
        move #<4-1,m5
        move #RxDataPhaseBuff,r4        ;r4/m4 points to phase buffer
        move #<RxDataScanLen-1,m4       ;to store differential phases
        move (r5)+
        move r5,X:<RxSymbolPhase        ;advance the symbol phase
        move X:<RxPipePtr,r5            ;r5/m5 points to the FFT phase pipe
        move X:<RxDataFirstBin,n0
        move #4*RxDataScanLen-1,m5
        move (r0)+n0                    ;advance to the first FFT bin we care about
        move #@rvb(DataCarrSepar,WindowLenLog),n0
        .loop #RxDataScanLen
         if SPY&&SPY_RxDataFFTidx
          clr a r0,b1
          .loop #WindowLenLog
            lsr b1
            rol a1
          .endl
          rep #8
            asl a
          jsr <SpyA
         endif
          move L:(r0)+n0,ab             ;take the FFT I/Q
          jsr <RaughNormIQ              ;and compute the phase
          jsr <Phase
          move Y:(r5),x0                ;read the phase 4 slices before this one
          sub x0,a a,Y:(r5)+            ;and compute the difference
          move a1,Y:(r4)+               ;save the phase difference
        .endl
        move r5,X:<RxPipePtr
      if SPY&&SPY_RxDataDiffPhases
        jsr <SpySync
        .if <cc>
          move #>30000<<8,a
          jsr <SpyA
          move X:<RxSymbolPhase,a
          rep #8
            asl a
          jsr <SpyA
          .loop #RxDataScanLen
            move Y:(r4),a
            jsr <SpyA
            move (r4)+
          .endl
        .endi
      endif
      if SPY&&(SPY_RxDataSelPhase>=0)
        move #<SPY_RxDataSelPhase,n5
        move X:<RxSymbolPhase,a
        rep #8
          asl a
        jsr <SpyA
        move Y:(r5+n5),a
        jsr <SpyA
      endif
      if SPY&&(SPY_RxDataSelDiffPhase>=0)
        move #<SPY_RxDataSelDiffPhase,n4
        move X:<RxSymbolPhase,a
        rep #8
          asl a
        jsr <SpyA
        move Y:(r4+n4),a
        jsr <SpyA
      endif

        move X:<RxPhaseDevPtr,r5
        move #4*RxDataPhaseDevLen-1,m5
        move #<RxDataPhaseDevMid-RxDataPhaseDev,n5
        move #FFTbuff,r0                ;here we re-use the FFTbuff
        move #DataCarriers/2-1,m0       ;because we don't need the FFT data
        move #<2,n4                     ;there anymore
        .loop #RxDataPhaseDevLen        ;for posible carrier positions
          move r4,n0                    ;we select the FFT bins
          .loop #DataCarriers/2
            move Y:(r4)+n4,a
            asl a
            move a1,Y:(r0)+
          .endl
          move n0,r4
          jsr <AverMS ; or AverMS_rej   ;and we compute the average
                                        ;phase shift and deviation
          move a,x0 #>RxDataAverWeight,y0       ;x0 = input
          mpy x0,y0,a #<1.0-RxDataAverWeight,y1
          mac x0,y0,a X:(r5+n5),x0              ;x0 = middle elem.
          mac x0,y1,a X:(r5),x1                 ;x1 = output
          macr -x1,y0,a #<4*RxDataAverWeight,y0 ;a = updated middle elem.
          mpy x0,y0,a a,X:(r5+n5)
          move #<1.0-4.0*RxDataAverWeight,y1
          macr x1,y1,a                          ;a = updated output
          move a,X:(r5)
          move b,x0 #>RxDataAverWeight,y0       ;x0 = input
          mpy x0,y0,b #<1.0-RxDataAverWeight,y1
          mac x0,y0,b Y:(r5+n5),x0              ;x0 = middle elem.
          mac x0,y1,b Y:(r5),x1                 ;x1 = output
          macr -x1,y0,b #<4*RxDataAverWeight,y0 ;a = updated middle elem.
          mpy x0,y0,b b,Y:(r5+n5)
          move #<1.0-4.0*RxDataAverWeight,y1
          macr x1,y1,b                          ;a = updated output
          move b,Y:(r5)+
          move (r4)+
        .endl
        move r5,X:<RxPhaseDevPtr
      if SPY&&SPY_RxDataPhaseDev
        jsr <SpySync
        .if <cc>
          clr a #<RxDataPhaseDevLen,n5
          jsr <SpyA
          move (r5)-n5
          .loop #RxDataPhaseDevLen
            move Y:(r5),a
            jsr <SpyA
            move (r5)+
          .endl
        .endi
      endif

        btst #0,X:<RxSymbolPhase
        .if <cc>
          move #RxDataPhaseBuff,r4
          move #<2,n4
          btst #1,X:<RxSymbolPhase
          .if <cs>              ;for odd symbols look through the odd carriers
            move (r4)+
          .endi
          move X:<RxDataSamplePtr,r2
          move #DataInterleave*2*RxDataSampleLen-1,m2
        if RxDataCorrAverShift
          move r4,n0    ;save r4
          rep #RxDataScanMargin
            move (r4)+n4
          .loop #DataCarriers/2 ;Y:(r0)/m0 still points to FFTbuff
            move Y:(r4)+n4,a
            asl a
            move a1,Y:(r0)+     ;dirty trick but the AverMS is hardwired
            move a1,Y:(r0)+     ;for a fixed number of data
          .endl
          jsr <AverMS           ;a = common phase shift
          asr a n0,r4           ;restore r4
          move a,y1
        else
          move #<0,y1
        endif
          .loop #RxDataSampleLen
            move Y:(r4)+n4,a    ;a = demodulated phase difference
            sub y1,a            ;subtract the common shift
            move a1,x0
            jsr <FastCosine     ;compute soft decision (modifies r0/m0/n0,x1)
            move a,x0 #>1.0/DataCarriers,y0
            mpyr -x0,y0,a       ;scale them down, so the FHT won't overflow
            move a,Y:(r2)+      ;by the way invert the sign
          .endl
          move #RxDataSampleLen,n2
          move r2,X:<RxDataSamplePtr
          move (r2)-n2
        if SPY&&SPY_RxDataSoftSamples
          jsr <SpySync
          .if <cc>
            .loop #RxDataSampleLen
              move Y:(r2),a
              jsr <SpyA
              move (r2)+
            .endl
            move (r2)-n2
          .endi
        endif
          move (r2)-n2
          btst #1,X:<RxSymbolPhase
          .if <cs>
            move #RxIntlvPatt0,r4
          .else
            move #RxIntlvPatt1,r4
          .endi
          move #<DataCarriers-1,m4      ;Y:(r4)/m4 = deinterleave pattern
          move #FFTbuff,r1              ;Y:(r1)/m1 = place for de-interleaved
          move #<DataCarriers-1,m1      ;blocks
          move X:<RxDecodeDevPtr,r5
          move #<2*2*RxDataScanMargin-1,m5
          move #<RxDecodeDevMid-RxDecodeDev,n5
          .loop #2*RxDataScanMargin
            move r2,r0
            .loop #DataCarriers
              move Y:(r2),x0
              move Y:(r4)+,n2
              move x0,Y:(r1)+
              move (r2)-n2
            .endl
            move r0,r2
          if SPY_RxDataDeIntlvSamples
           btst #1,X:<RxSymbolPhase
           .if <cs>
            jsr <SpySync
            .if <cc>
              .loop #DataCarriers
                move Y:(r1),a
                jsr <SpyA
                move (r1)+
              .endl
            .endi
           .endi
          endif
            jsr <FHT
            jsr <RxDecodeChar   ;r0 = char., a = noise, b = peak
            move r0,Y:(r5)
            move a,Y:(r5+n5)
            move a,x0 #>RxDataAverWeight,y0       ;x0 = input
            mpy x0,y0,a #<1.0-RxDataAverWeight,y1
            mac x0,y0,a X:(r5+n5),x0              ;x0 = middle elem.
            mac x0,y1,a X:(r5),x1                 ;x1 = output
            macr -x1,y0,a #<4*RxDataAverWeight,y0 ;a = updated middle elem.
            mpy x0,y0,a a,X:(r5+n5)
            move #<1.0-4.0*RxDataAverWeight,y1
            macr x1,y1,a (r2)+                    ;a = updated output
            move a,X:(r5)+
          .endl
          move r5,X:<RxDecodeDevPtr
          btst #1,X:<RxSymbolPhase
          .if <cc>
         if SPY&&SPY_RxDataDecodeDev
            jsr <SpySync
            .if <cc>
              .loop #2*2*RxDataScanMargin
                move X:(r5),a
                jsr <SpyA
                move (r5)+
              .endl
            .endi
         endif
            move X:-(r5),a
            move r5,r0
            .loop #2*2*RxDataScanMargin-1
              move X:-(r5),x0
              cmp x0,a
              tgt x0,a r5,r0
            .endl
          move r0,b
          move r5,x0
          sub x0,b a,X:<RxDecodeMinNoise
          move b,X:<RxDecodeBestCarr
          if SPY&&SPY_RxDataDecodeChar
            move X:<RxDecodeMinNoise,a
            jsr <SpyA
            move Y:(r0),a
            rep #8
              asl a
            jsr <SpyA
          endif
          if !SPY
            move r0,r5
            move Y:(r0),a
           if RxCharSelThres>0
            move #<RxCharSelThres,x0
            move Y:(r5+n5),b
            cmp x0,b #>RxBadCharFill,x1
            tgt x1,a
           endif
            jsr <OutputChar
          endif
          .endi
        .endi
      if (!SPY)&&RxReport
        btst #RxSta_QualityReq,X:<RxState
        .if <cs>
          jsr <RxQualityReport
          bclr #RxSta_QualityReq,X:<RxState
        .endi
      endif

        move X:<RxSymbolPhase,a ;the DCD logic, timing and frequency
        tst a           ;correction we only do at the proper symbol time
        jne <ProcessLoop

        move X:<RxPhaseDevPtr,r5
        move #4*RxDataPhaseDevLen-1,m5
        move #<RxDataPhaseDevLen,n5             ;here we look for the
        move #<2*2*RxDataScanMargin+1-1,m0      ;lowest phase deviation
        move #<0,r0     ;r0 will count carriers
        move (r5)-n5
        move #<2*RxDataPhaseDevLen+1,n5
        move Y:(r5),a
        move Y:(r5+n5),x0
        add x0,a r0,n0  ;a = minimum level, n0 = carrier where minimum
        move r5,y1      ;y1 = where in memory
        .loop #2*2*RxDataScanMargin+1 ;scan through carriers
          move Y:(r5+n5),b
          move Y:(r5),x0
          add x0,b r5,x0
          cmp a,b (r5)+
          .if <lt>            ;if lower dev. found then
            tfr b,a r0,n0     ;register its value and position
            move x0,y1
          .endi
          move (r0)+
        .endl
        move y1,r5           ;a = phase Mean Square deviation at best point
        asr a
        move X:(r5),b
        move X:(r5+n5),x0
        add x0,b a,X:<RxSignalMinDev
        asr b #RxDataDropThres,x1   ;b = phase offset at lowest deviation
        move b,X:<RxTmpI
      if SPY&&SPY_RxDataScanMin
        move X:<RxSymbolPhase,a ;symbol phase
        rep #8
          asl a
        jsr <SpyA
        move n0,a               ;best carrier position
        move #>2*RxDataScanMargin,x0
        sub x0,a
        rep #8
          asl a
        jsr <SpyA
        move X:<RxTmpI,a        ;phase offset at minimum
        jsr <SpyA
        move X:<RxSignalMinDev,a ;minimum phase dev.
        jsr <SpyA
      endif
        ;a = minimum deviation, b = freq. offset, x1 = threshold for DCD
        cmp x1,a X:<RxAcceptCount,a
        move #<RxDataAverWeight,y0
        .if <lt>
          add y0,a
          bset #RxSta_NowDCD,X:<RxState
        .else
          sub y0,a
          bclr #RxSta_NowDCD,X:<RxState
        .endi
        move a,X:<RxAcceptCount
        move #<-0.75,x0
        cmp x0,a X:<RxCarFreq,a
        jle <RxCarrierLost
        btst #RxSta_NowDCD,X:<RxState
        jcc <ProcessLoop        ;no corrections if signal too weak
        move b,x0               ;freq. offset tracking
        move #>(1.0/SymbolSepar/2)*RxDataFreqTrackWeight,x1
        move X:<RxCarFreq,a
        macr -x0,x1,a X:<RxTimeCorrHold,b
        sub y0,b a,X:<RxCarFreq
        move b,X:<RxTimeCorrHold
        jge <ProcessLoop        ;avoid timing correction if there was one recently
        move n5,n4              ;look at the timing shift
        move Y:(r5+n5),x0 ;Y:(r5)/m5/n5 still points to phase MS deviations
        move #<RxDataPhaseDevLen,n5
        move m5,m4
        lua (r5)-n5,r4
        move Y:(r5),b
        add x0,b Y:(r4+n4),x0 ;b = lowest deviation -> x1 (not used here)
        move Y:(r4),a         ;
        lua (r5)+n5,r4
        add x0,a b,x1         ;a = deviation one slice before -> y0
        move Y:(r4+n4),x0
        move Y:(r4),b
        add x0,b a,y0         ;b = deviation one slice after -> y1
        cmp a,b  b,y1         ;compare
        move #>1,a
        .if <gt>
          neg a
        .endi
        move #>RxSliceSepar,b
        add b,a #<$7F,b
        move a,X:<RxSliceSeparDeci
        move b,X:<RxTimeCorrHold

        jmp <ProcessLoop

RxDecodeChar    ;Y:(r1)/m1 = the FHT result
        move Y:(r1)+,x0 ;first element
        tfr x0,b        ;b = supposed maximum, r0 = maximum position+2
        mpy x0,x0,a Y:(r1)+,x0  ;a = square sum, x0 = second element
        move m1,m0
        move r1,r0
        .loop #DataCarriers-1
          mac x0,x0,a x0,y0
          cmpm x0,b Y:(r1)+,x0
          tlt y0,b r1,r0
        .endl
        move (r1)-
        move r1,n0
        move (r0)-
        tst b (r0)-
        .if <mi>
          bchg #6,r0    ;here we assume DataCarriers=64
        .endi
        move b,x0
        macr -x0,x0,a (r0)-n0
        rts                     ;r0 = decoded symbol (7-bit ASCII)
                                ;a = decoder noise
                                ;b = highest element
                                ;r1 moves but returns to original position
RxCarrierLost
       if (!SPY)&&RxReport
         move #$FFFF,m1
         move #RxLockLostMsg,r1
         jsr <PutString
       endif
        jmp <RxStartIdle

      if (!SPY)&&RxReport
RxQualityReport
        move #$FFFF,m1
        move X:<RxDecodeFirstBin,a
        move #>FirstDataCarr,x0
        sub x0,a X:<RxDecodeBestCarr,b
        move a,x0 #>2*RxDataScanMargin,y0
        sub y0,b #>DataCarrSepar,y0
        mpy x0,y0,a X:<RxCarFreq,x1
        asr a
        move a0,a
        add b,a #>2.0/WindowLen/2,y0
        move a,x0
        mpy x0,y0,a #>1.0/SymbolSepar/2,y0
        move a0,a
        macr -x0,y0,a #RxQualMsg_freq,r1
        move a,x0
        jsr <FormatFreq_sdd
        move X:<RxSignalMinDev,a
        jsr <Log2 ;modifies r0,x,b
        neg a #RxQualMsg_SNR,r1
        rep #6
          asr a
        move a,x0
        jsr <FormatPosLog2frac_ddd_dB
        move X:<RxDecodeMinNoise,a
        jsr <Log2 ;modifies r0,x,b
        neg a #RxQualMsg_FEC,r1
        rep #6
          asr a
        move a,x0
        jsr <FormatPosLog2frac_ddd_dB
        move #RxQualMsg,r1
        jsr <PutString
        rts
      endif

;==============================================================================
;State machine routines for the transmitter

TxAbort
       if (!SPY)&&TxReport
         move #$FFFF,m1
         move #TxAbortMsg,r1
         jsr <PutString
       endif
        move #>TxCharBuff,x0    ;clear the character buffer
        move x0,X:<TxCharBuffRd
        move x0,X:<TxCharBuffWr
        bclr #TxSta_PTTabortReq,X:<TxState ;clear the abort request
        bclr #TxSta_PTTonReq,X:<TxState    ;and the on request too

TxStartIdle
        PTT clr                         ;turn off the PTT
        move #>TxDoIdle,x0
        move x0,X:<TxStateRoutine
        move #>2*TxMinIdle,x0   ;obligatory silence period
        move x0,X:<TxStateCounter
      if SPY
;        jsr <Rand48             ;randomly shift around the symbol timing
;        move a0,x0              ;to see the receiver's response
;        eor x0,a1 #>DecimateRatio*SymbolSepar-1,x0
;        and x0,a1               ;make a random number
;        lsl a1                  ;multiply by 4
;        lsl a1 X:<TxCodecPtr,r2 ;move the tx pointer back by this distance
;        move a1,n2
;        move #CodecBuffLen*4-1,m2
;        nop
;        move (r2)-n2
;
;        move r2,X:<TxCodecPtr
;        move X:<TxCodecPtr,r2 ;move the tx pointer back by:
;        move #4*((DecimateRatio*SymbolSepar)/8),n2
;        move #CodecBuffLen*4-1,m2
;        nop
;        move (r2)-n2
;        move r2,X:<TxCodecPtr
      endif
TxDoIdle                        ;send nothing (silence)
        jset #TxSta_PTTabortReq,X:<TxState,TxAbort
        move X:<TxStateCounter,a ;when counter expires
        tst a #>1,x0            ;check if any data for transmition
      if SPY
        jle <TxStartSync
      else
        jle <TxCheckPTTreq
      endif
        sub x0,a
        move a,X:<TxStateCounter
        jmp <TxDoSilence

      if !SPY
TxCheckPTTreq
        jset #TxSta_PTTonReq,X:<TxState,TxStartSync
      endif ; of !SPY

      jmp <TxDoSilence

TxStartSync
       if (!SPY)&&TxReport
         move #$FFFF,m1
         move #TxStartMsg,r1
         jsr <PutString
       endif
        bclr #TxSta_Symbol,X:<TxState
        move #TxInitVectors,r4  ;copy the initial vectors
        move #<DataCarriers-1,m4
        move #TxWorkVectors,r5  ;to the work area
        move m4,m5
        .loop #DataCarriers
          move L:(r4)+,ab
          move ab,L:(r5)+
        .endl
        PTT set                 ;turn on the PTT
        move #>TxDoSync,x0      ;set the state routine
        move x0,X:<TxStateRoutine
        move #>2*TxSyncLen,x0   ;and the state counter
        move x0,X:<TxStateCounter
        clr a #TxIntlvPipe,r1   ;clear the Tx interleaver
        move #DataInterleave*WordsPerBlock-1,m1
        .loop #DataInterleave*WordsPerBlock
          move a,Y:(r1)+
        .endl
        move r1,X:<TxIntlvPtr   ;restart the interleaver pointer
        move #TxStartCharSeq,r1 ;set message pointer
        move r1,X:<TxMsgPtr     ;to the startup sequence
TxDoSync                        ;send the startup sequence
        jset #TxSta_PTTabortReq,X:<TxState,TxAbort
        move X:<TxStateCounter,a
        tst a #>1,x0            ;if enough symbols then proceed with
        jle <TxStartData        ;the data phase
        sub x0,a #TxVect,r5                 ;L:r5/m5 = Tx vector
        move a,X:<TxStateCounter
        move #<DataCarriers-1,m5
        bchg #TxSta_Symbol,X:<TxState   ;if we are at odd symbols then only
        jcs <TxOutputOdd                ;send out the odd part of the TxVect
        move X:<TxMsgPtr,r1             ;otherwise made a new character
        move #<TxSyncLen-1,m1           ;from the startup message
      if TxReport
        move Y:(r1),a1                ;echo the character back
        jsr <OutputChar
      endif
        move #<TxDataBlock,r0
        move #<WordsPerBlock-1,m0
        move Y:(r1)+,a1                 ;a1 = the new character
        move r1,X:<TxMsgPtr
        jsr <TxEncodeChar               ;encode char. into a Walsh vector
        jsr <TxIntlvBlock               ;pass through the interleaver
        jsr <TxEncodeBlock              ;encode onto phases in TxVect
        jmp <TxOutputEven               ;output the even part of TxVect
                                        ;odd part will follow at next pass
TxStartData
        bclr #TxSta_PTTonReq,X:<TxState ;clear the PTT-ON request
        move #>TxDoData,x0
        move x0,X:<TxStateRoutine
      if SPY                    ;in SPY mode we send a fixed number of chars
        move #>2*SPY_TxRandSymbols,x0
        move x0,X:<TxStateCounter
      endif
TxDoData
      if SPY                    ;in SPY mode count the chars.
        move X:<TxStateCounter,a
        tst a #>1,x0
        jle <TxStartPostData
        sub x0,a
        move a,X:<TxStateCounter
      else ; of SPY             ;in normal mode check for PTT-OFF request
        jset #TxSta_PTTabortReq,X:<TxState,TxAbort
        btst #TxSta_PTToffReq,X:<TxState
        .if <cs>
          move X:<TxCharBuffRd,a1       ;stop transmitting data only
          move X:<TxCharBuffWr,x0       ;when the TxCharBuff is empty
          eor x0,a1
          jeq <TxStartPostData
        .endi
      endif ; of SPY
        move #TxVect,r5                 ;L:r5/m5 = Tx vector
        move #<DataCarriers-1,m5
        bchg #TxSta_Symbol,X:<TxState   ;if we are at odd symbols then only
        jcs <TxOutputOdd                ;send out the odd part of the TxVect
      if SPY
;        jsr <Rand48                     ;make a random character
;        move a0,x0
;        eor x0,a1                       ;a1 = the random character
      else
        move X:<TxCharBuffRd,a1         ;get the next user character
        move #TxCharBuffLen-1,m1        ;from the TxCharBuff
        move X:<TxCharBuffWr,x0
        eor x0,a1 a1,r1
        .if <ne>
         if TxReport
          move Y:(r1),a1                ;echo the character back
          jsr <OutputChar
         endif
          move Y:(r1)+,a1
          move r1,X:<TxCharBuffRd
        .else                           ;if buffer empty, send NL
          move #<NL,a1
        .endi
      endif
        move #<TxDataBlock,r0
        move #<WordsPerBlock-1,m0
        jsr <TxEncodeChar               ;encode a1 into a Walsh vector Y:(r0)/m0
        jsr <TxIntlvBlock               ;pass through the interleaver
        jsr <TxEncodeBlock              ;encode onto phases in TxVect
        jmp <TxOutputEven               ;output the even part

TxStartPostData
        bclr #TxSta_PTToffReq,X:<TxState      ;clear the PTT-OFF request
        move #>TxDoPostData,x0
        move x0,X:<TxStateRoutine
        move #<2*(TxStopLen+DataInterleave+1),a
        move a,X:<TxStateCounter
        move #TxStopCharSeq,r1
        move r1,X:<TxMsgPtr
TxDoPostData
      if !SPY
        jset #TxSta_PTTabortReq,X:<TxState,TxAbort
        jset #TxSta_PTTonReq,X:<TxState,TxStartData ;if there is a new request for the PTT
                                       ;we append the new data
      endif                            ;to the ongoing transmition
        move X:<TxStateCounter,a
        tst a #<1,x0
        jle <TxStartJam
        sub x0,a #TxVect,r5                 ;L:r5/m5 = Tx vector
        move a,X:<TxStateCounter
        move #<DataCarriers-1,m5
        bchg #TxSta_Symbol,X:<TxState   ;if we are at odd symbols then only
        jcs <TxOutputOdd                ;send out the odd part of the TxVect
        move X:<TxMsgPtr,r1             ;send the stop message
        move #<TxStopLen+DataInterleave+1-1,m1
      if TxReport
        move Y:(r1),a1                ;echo the character back
        jsr <OutputChar
      endif
        move #<TxDataBlock,r0
        move Y:(r1)+,a1
        move #<WordsPerBlock-1,m0
        move r1,X:<TxMsgPtr
        jsr <TxEncodeChar
        jsr <TxIntlvBlock
        jsr <TxEncodeBlock
        jmp <TxOutputEven

TxStartJam
        move #>TxDoJam,x0
        move x0,X:<TxStateRoutine
        move #<2*TxJamLen,x0
        move x0,X:<TxStateCounter
TxDoJam                                 ;send the jamming sequence
        jset #TxSta_PTTabortReq,X:<TxState,TxAbort
        move X:<TxStateCounter,a
        tst a #<1,x0
        jle <TxStartBuffFlush
        sub x0,a #TxVect,r5                 ;L:r5/m5 = Tx vector
        move a,X:<TxStateCounter
        move #<DataCarriers-1,m5
        bchg #TxSta_Symbol,X:<TxState   ;if we are at odd symbols then only
        jcs <TxOutputOdd                ;send out the odd part of the TxVect
        jsr <TxEncodeRandomBlock
        .loop #DataCarriers     ;additionally shift the vectors by 90 degrees
          move L:(r1),ba        ;for maximum phase error and quick DCD loss
          neg b a,X:(r1)        ;at the receiver
          move b,Y:(r1)+
        .endl
        jmp <TxOutputEven

TxStartBuffFlush
      if (!SPY)&&TxReport
        move #$FFFF,m1
        move #TxStopMsg,r1
        jsr <PutString
      endif
        move #>TxDoBuffFlush,x0
        move x0,X:<TxStateRoutine
        move #<2*(CodecBuffLen/(SymbolSepar*DecimateRatio)+2),x0
        move x0,X:<TxStateCounter
TxDoBuffFlush                           ;let the CODEC's buffer flush
        jset #TxSta_PTTabortReq,X:<TxState,TxAbort
        move X:<TxStateCounter,a        ;the output samples before
        tst a #<1,x0                    ;we turn off the PTT
        jle <TxStartIdle                ;otherwise we will cut the transmition
        sub x0,a                        ;before all the audio samples
        move a,X:<TxStateCounter        ;come out to the TRX
        jmp <TxDoSilence

;TxEncodeZeroBlock                      ;we don't need this anymore
;        clr a #<TxDataBlock,r0
;        move #<WordsPerBlock-1,m0
;        .loop #WordsPerBlock
;          move a,Y:(r0)+
;        .endl
;        jmp <TxEncodeWord

;Encode a random word
TxEncodeRandomBlock                     ;r5/m5 points to TxVect
        move #<TxDataBlock,r0
        move #<WordsPerBlock-1,m0
        .loop #WordsPerBlock            ;create new random word
          jsr <Rand48                   ;a1:a0 = a new random number
          move a0,x0                    ;mix the two random words
          eor x0,a1                     ;for more "randomness"
          move a1,Y:(r0)+               ;(the Rand48 is a very primitive stuff)
        .endl
;encode word given by Y:(r0)/m0
TxEncodeBlock                           ;Y:(r0)/m0= TxDataBlock
        move #TxWorkVectors,r1          ;L:r1/m1 = symbol phases
        move #<DataCarriers-1,m1        ;r5/m5 points to TxVect
        move #SinTable+(SymbolSepar*FirstDataCarr)%WindowLen,r2
        move #WindowLen-1,m2            ;r2/m2/n2 for correcting FFT phases
        move #(SymbolSepar*DataCarrSepar)%WindowLen,n2
        .loop #DataCarriers             ;
          move X:(r1),x1                ;x1:x0 = phase
          move #TxDataAmpl,y0           ;y0 = data amplitude
          mpyr x1,y0,a Y:(r1),x0        ;apply phase to the tone's amplitude
          mpyr x0,y0,b a,X:(r5)         ;put the phased tone to the TxVect
          move X:(r2),y1                ;y1:y0 = phase correction
          mpy x1,y1,a Y:(r2)+n2,y0      ;apply phase correction
          macr -x0,y0,a b,Y:(r5)+
          mpy x1,y0,b a,X:(r1)
          macr x0,y1,b
          move b,Y:(r1)+
        .endl
        move #>1,x1             ;single bit mask
        .loop #DataCarriers
          move Y:(r0),a1        ;get the mask
          and x1,a1             ;apply to the word
          .if <eq>              ;if bit==0 then flip the phase
            move X:(r1),a
            neg a Y:(r1),b
            neg b a,X:(r1)
            move b,Y:(r1)
          .endi                 ;otherwise (bit==1) don't touch the phase
          move x1,a1
          lsl a1 (r1)+          ;rotate the mask to the left
          .if <cs>              ;if the 1-bit has come out
            rol a1 (r0)+        ;then rotate once more (so the 1 comes back in)
          .endi                 ;and increment the word pointer
          move a1,x1            ;store back the mask in x1
        .endl
        rts

TxOutputEven                    ;r5/m5 points to the TxVect
        clr a #FFTbuff,r4       ;set all FFT bins to 0
        move #WindowLen-1,m4
        .loop #WindowLen
          move a,L:(r4)+
        .endl
        move #FFTbuff+FirstDataCarr,r4
        move #<2*DataCarrSepar,n4
        move #<2,n5
        .loop #(DataCarriers+1)/2 ;copy even vectors
          move Y:(r5),b
          neg b X:(r5)+n5,a     ;negate im. part for the IFFT (?)
          move ab,L:(r4)+n4
        .endl
        jmp <TxDoFFT

TxOutputOdd
        clr a #FFTbuff,r4       ;set all FFT bins to 0
        move #WindowLen-1,m4
        .loop #WindowLen
          move a,L:(r4)+
        .endl
        move #FFTbuff+FirstDataCarr+DataCarrSepar,r4
        move #<2*DataCarrSepar,n4
        move #<2,n5
        move (r5)+
        .loop #DataCarriers-((DataCarriers+1)/2) ;copy odd vectors
          move Y:(r5),b
          neg b X:(r5)+n5,a     ;negate im. part for the IFFT (?)
          move ab,L:(r4)+n4
        .endl
        jmp <TxDoFFT

;------------------------------------------------------------------------------
;CODEC's buffer synchronization to the sample stream

CheckSampleBlock        ;a = how many samples we want
        asl a           ;r2 = where we wait for the block to start
        asl a #>2,x1
        add x1,a
        move a,x1
CheckSampleBlock_1
        move r7,a
        move r2,x0
        sub x0,a
        jpl <CheckSample_cmp
          move #>CodecBuffLen*4,x0
          add x0,a
CheckSample_cmp
        cmp x1,a
        rts             ;on output: a,x0,x1 are modified
                        ;If carry=0 => we are sure that:
                        ;1. there are at least A new input samples at X:(r2)
                        ;   to be read
                        ;2. there are at least A locations at Y:(r2) to
                        ;   be written with output samples


WaitSampleBlock         ;wait for a block of samples from the CODEC
                        ;on input: a = how many samples we want
                        ;          r2 = where the block starts
        jsr <CheckSampleBlock
        jcc <WaitSample_ret
WaitSample_loop
        wait            ;wait (at low power) until an interrupt comes
        jsr <CheckSampleBlock_1
        jcs <WaitSample_loop
WaitSample_ret
        rts             ;on output: a,x0,x1 are modified
                        ;We are sure that:
                        ;1. there are at least A new input samples at X:(r2)
                        ;   to be read
                        ;2. there are at least A locations at Y:(r2) to
                        ;   be written

;------------------------------------------------------------------------------
;Trigonometric functions

;Fast cosine for computing soft decisions - takes only 8 bits of the phase
FastCosine              ;x0 = phase/angle (-1.0 = -PI, +1.0 = +PI)
        move #>$80,x1
        mpy x0,x1,a #>$FF,x1
        and x1,a1 #>$140,r0
        move x1,m0
        move a1,n0
        nop
        move Y:(r0+n0),a
        rts             ;a = cosine(x0) by table lookup
                        ;r0/m0/n0,x1 are modified

PI      equ     3.14159265358979323846

;this routine computes a cosine/sine pair using the sine ROM
;with a second order (linear+quadrature) approximation between table points
IQ                              ;x0 = angle ( -1 = -PI, +1 = +PI)
;        ori #%00000100,omr      ;enable the sine ROM table
        move #>$80,x1   ;shift out 8 most significant bits
        mpy x0,x1,a  #>$FF,x0
        move x0,m0
        and x0,a     #>$100,x0
        or x0,a      #<$40,n0
        move a1,r0      ;put the 8 most significant bits into r0 with offset = $100
        move a0,y0      ;save the remaining bits in y0
        jclr #23,y0,SinTable_lev2
          move (r0)+
SinTable_lev2
        move Y:(r0+n0),x0       ;x0 = coarse cosine
        move Y:(r0),x1          ;x1 = coarse sine
        mpyr x1,y0,a  #PI/256.0,y1
        tfr x0,a  a,x1
        macr -x1,y1,a           ;a = fine cosine
        mpyr x0,y0,b  Y:(r0),x1
;        andi #%11111011,omr     ;disable the sine ROM table
        tfr x1,b  b,x1
        macr x1,y1,b  #PI*PI/2.0/65536.0,y1  ;b = fine sine
        mpyr y0,y0,a  a,x0
        move a,y0
        mpyr y0,y1,a
        tfr x0,a  a,y1
        macr -x0,y1,a  b,x1     ;a = super fine cosine
        macr -x1,y1,b           ;b = super fine sine
        rts                     ;x,y are modified
                                ;r0,m0,n0 are modified
                                ;maximum error is about 0.7E-6
                                ;execution time 4+64+4 clock cycles
                                ;including "jsr <IQ_lev2" and "rts"

RaughNormIQ             ;scale an I/Q pair such that the larger number
                        ;is within the range +/-0.5..1.0
        cmpm a,b        ;which number is greater ?
        jge <NormAftB   ;if b then follow the normalization according to it
        tst a           ;otherwise follow a
        jes <RightA     ;if too big (extension in use) the shift to the right
        jnr <DoneA      ;if normalized then we are done
LeftA     asl b         ;otherwise shift left both numbers
	  asl a
        jnn <LeftA      ;until A is normalized
        rts
RightA    asr b         ;if extension set then shift both number to the right
	  asr a
        jes <RightA     ;until a is normalized
DoneA   rts

NormAftB                ;same stuff for b
	tst b
        jeq <DoneB
        jes <RightB
        jnr <DoneB
LeftB	  asl a
	  asl b
	jnn <LeftB
        rts
RightB	  asr a
	  asr b
	jes <RightB
DoneB   rts


;The following routine computes the phase of the given I/Q pair
;this routine is not very precise: maximum error is 2E-4 (2E-4/PI radians)
;the error could be made lower but I required that the calculation
;is exact for PI/4, PI/2, PI*3/4, etc.
Phase      ;a = I-part (24 bits), b = Q-part (24 bits)
           ;-1.0 <= a < 1.0, -1.0 <= b < 1.0
        tst b #<0,x0    ;if Q-part negative => turn by PI
        .if <mi>
          neg a #<%10000000,x0
          neg b
        .endi
        tst a           ;if I-part negative => turn by -PI/2
        .if <mi>
          bset #22,x0
          tfr a,b b,a   ;swap I with Q
          neg b         ;negate the new Q
        .endi
        add b,a                 ;add Q to I (what if I+Q >= 1.0 ?)
        asr a                   ;I+Q /= 2 in case it is > 1
        asr b a,x1              ;Q /= 2 because of the above, x1 = I+Q
        andi #$FE,ccr           ;divive Q by (I+Q)
        rep #24
          div x1,b
        tfr x0,a b0,x0          ;x0 = Q div (I+Q), a=approx angle
        tfr x0,b #0.5,x0        ;b = Q/(I+Q)
        sub x0,b #0.25,x0       ;b = Q/(I+Q) - 1/2 =: X
        add x0,a b,x1 #0.63361654,y1 ;angle += 0.25, x1 = X
        mac x1,y1,a x1,x0       ;angle += F1*X, x0 = X
        mpyr x0,x0,b #-0.73728217,y1 ;b = X*X
        mpyr x1,y1,b b,x0       ;b = F3*X, x0 = X*X
        move b,y1
        mac x0,y1,a x0,y0       ;angle += F3*X * X*X, y0 = X*X
        mpyr x0,x0,b #0.85568791,y1 ;b = (X*X)*(X*X)
        mpyr x1,y1,b b,x0       ;b = F5*X, x0 = X*X*X*X
        move b,y1
        mac x0,y1,a             ;angle += F5*X * X*X*X*X
        mpyr x0,y0,b #-0.17769569,y1 ;b = (X*X*X*X)*(X*X)
        mpyr x1,y1,b b,x0       ;b = F5*X, x0 = X*X*X*X*X*X
        move b,y0
        macr x0,y0,a            ;angle += F5*X * X*X*X*X*X*X
        rts             ;a = the phase (-1.0 => -PI, 1.0 => PI)
                        ;b,x,y are modified
                        ;execution time is about 130 cycles (65 instructions)


;------------------------------------------------------------------------------
;Random number generator

Rand48                          ;a simple random number generator
        move L:<Rand,a
        asl a L:<RandInc,x
        adc x,a
        move a10,L:<Rand
        rts                     ;returns a pseudo-random number in a
                                ;modifies x

;------------------------------------------------------------------------------
;FEC encoders and decoders


;------------------------------------------------------------------------------

;------------------------------------------------------------------------------
;The interleaver for the transmitter

TxIntlvBlock            ;Y:(r0)/m0 = input block
;        rts ;*** DEBUG *** to turn the interleaver off
        move X:<TxIntlvPtr,r1
        move #<WordsPerBlock*DataInterleave-1,m1
        .loop #WordsPerBlock    ;copy the new word into the interleaver pipe
          move Y:(r0)+,a1
          move a1,Y:(r1)+
        .endl
        move r1,X:<TxIntlvPtr   ;update the pipe pointer
        clr b #>1,x1            ;b = word being formed, x1 = data mask
        move #TxIntlvPatt,r2
        move #<BitsPerBlock-1,m2
        .loop #BitsPerBlock     ;pick up the proper bits from the pipe
          lsr b1 Y:(r1),a1
          and x1,a1 Y:(r2)+,n1  ;pick the bit according to mask,
          .if <ne>
            bset #23,b1         ;put the bit into the word
          .endi
          move x1,a1            ;shift the mask
          lsl a1 (r1)+n1        ;get next interleaver step
          .if <cs>              ;if 24-bits are formed
            rol a1 (r1)+        ;then advance the pointer
            clr b b1,Y:(r0)+    ;save the word
          .endi
          move a1,x1
        .endl
       if (WordsPerBlock*24-BitsPerBlock)>0
        rep #(WordsPerBlock*24-BitsPerBlock)
          lsr b1
       endif
        move b1,Y:(r0)+
        rts             ;Y:(r0)/m0 = output block
                        ;r1/m1/n1, r2/m2, a,b,x1 are modified

TxEncodeChar            ;a1 = user code to be sent out (should be a 7-bit ASCII)
        move #>$7F,x1   ;Y:(r0)/m0 = where to place the encoded data block
        and x1,a1 #<$01,b       ;apply the 7-bit mask
        btst #6,a1
        .if <cs>
          bclr #6,a1
          neg b
        .endi
        clr a a1,n1
        move #FHTbuff,r1
        move #<BitsPerBlock-1,m1
        .loop #BitsPerBlock     ;clear all elements
          move a,Y:(r1)+
        .endl
        move b,Y:(r1+n1)        ;set the N-th element to non-zero
        jsr <IFHT               ;perform Inverse Hadamard Transform
        move #<$80,a
        .loop #BitsPerBlock
          btst #23,Y:(r1)       ;if negative element -> data bit = 1
          ror a1
          .if <cs>
            move a1,Y:(r0)+
            move #<$80,a
          .endi
          move (r1)+
        .endl
       if (WordsPerBlock*24-BitsPerBlock)>0
        rep #(WordsPerBlock*24-BitsPerBlock)
          lsr a1
       endif
        move a1,Y:(r0)+
        rts             ;r1/m1/n1,a,b,x are modified
                        ;Y:(r0)/m0 = data block

;------------------------------------------------------------------------------




;------------------------------------------------------------------------------

      if RxAGC

RxGainUp                        ;increase the CODEC's input gain (both channels)
        clr a #>$0F0F00,x0      ;r2/m2 should point to the CODEC's buffer
        move Y:(r2),a1          ;get the CODEC's input control word
        and x0,a                ;extract the gain bits
	cmp x0,a  #>$010100,x0  ;already maximum ?
        jeq <RxGainChange_abort ;if so then abort this attempt
	add x0,a  #>$F0F000,x0  ;if not, increment the gain by 1
	move a1,x1
	move Y:(r2),a1          ;and reload all the control words
	and x0,a  n2,x0         ;in the output buffer
	or x1,a  #<4,n2         ;make n2=4 for a moment
        .loop #CodecBuffLen
	  move a1,Y:(r2)+n2
	.endl
	move x0,n2              ;restore n2
        andi #$FE,ccr           ;clear carry
        rts                     ;modifies a,x0,x1

RxGainDown                      ;decrease the CODEC's input gain (both channels)
        clr a #>$0F0F00,x0      ;r2/m2 should point to the CODEC's buffer
        move Y:(r2),a1          ;get the CODEC's input control word
        and x0,a  #>$010100,x0  ;extract the gain bits
        sub x0,a  #>$F0F000,x0  ;attempt to decrease the gain
        jcs <RxGainChange_abort ;jump if overflow
        move a1,x1
        move Y:(r2),a1          ;reload all the input control words
        and x0,a  n2,x0         ;in the buffer with the new input gain
        or x1,a  #<4,n2         ;n2=4 for a moment
        .loop #CodecBuffLen
          move a1,Y:(r2)+n2
        .endl
        move x0,n2              ;restore n2
        andi #$FE,ccr           ;clear carry
        rts                     ;modifies a,x0,x1

RxGainChange_abort
        ori #$01,ccr            ;set carry to indicate that we couldn't
        rts                     ;change the gain

      endif ; of RxAGC

;------------------------------------------------------------------------------

      if !SPY

OutputChar      ;send a character in a1 to terminal
        move a1,x0
        move #<CR,a1        ;is this CR ?
        eor x0,a1
        .if <eq>            ;if it is then send LF afterwards
           putc
           move #>LF,x0
           putc
        .else
           putc
        .endi
        rts     ;a,b,x,y,r0/m0 can be modified

CheckInput
        move X:<TxCharBuffWr,r1
        move #TxCharBuffLen-1,m1
CheckInput_loop
        lookc 0
        jcs <CheckInput_ret
        move #<ReceiveSwitch,a1
        eor x0,a1 x0,Y:(r1)+
        .if <eq>
          bset #TxSta_PTToffReq,X:<TxState
          jmp <CheckInput_loop
        .endi
        move #<TransmitSwitch,a1
        eor x0,a1 
        .if <eq>
          bset #TxSta_PTTonReq,X:<TxState
          jmp <CheckInput_loop
        .endi
        move #<RxAbortSwitch,a1
        eor x0,a1 
        .if <eq>
          bset #RxSta_AbortReq,X:<RxState
          jmp <CheckInput_loop
        .endi
        move #<RxQualitySwitch,a1
        eor x0,a1 
        .if <eq>
          bset #RxSta_QualityReq,X:<RxState
          jmp <CheckInput_loop
        .endi
        move #<TxAbortSwitch,a1
        eor x0,a1 X:<TxCharBuffRd,x0
        .if <eq>
          bset #TxSta_PTTabortReq,X:<TxState
          jmp <CheckInput_loop
        .endi
        move r1,a1
        eor x0,a1 #>BELL,x0
        jne CheckInput_loop
        move (r1)-
        putc
        jmp <CheckInput_loop
CheckInput_ret
        move r1,X:<TxCharBuffWr
        rts

      endif ; of !SPY

;------------------------------------------------------------------------------

      if (!SPY)&&RxReport

PutString                       ;put out a non-packed string at Y:(r1)/m1
        move Y:(r1)+,a
        tst a a,x0
        jeq <PutString_ret
        putc
        jmp <PutString
PutString_ret
        rts                     ;r1 advanced by the number of chars. processed
                                ;a,b,x,y,r0/m0 is modified due to putc

FormatSign                      ;a = the number with the sign
        tst a #<'+',b1          ;Y:(r1)/m1 = where to put the sign
        .if <mi>
          neg a #<'-',b1
        .endi
        move b1,Y:(r1)+
        rts                     ;a = the number without the sign (positive)
                                ;r1 advanced by 1

FormatDot
        move #<'.',b1           ;format the dot
        move b1,Y:(r1)+
        rts                     ;b1 only is modified and r1 advances

FormatFraction                  ;format the fractional part of A in decimal digits
                                ;must be: 1.0 > A > 0.0
        .loop n1                ;n1 = number of digits to be send
          asl a                 ;Y:(r1)/m1 = where to store the ASCII form
          tfr a,b
          asl a
          asl a
          add b,a               ;multiply A by 10
          tfr a,b
          asl a                 ;one more shift and save the number
          move a2,x0            ;now a2 contains the decimal digit
          move #>'0',a          ;convert to ASCII digit
          add x0,a #<0,b2
          tfr b,a a,Y:(r1)+     ;and put it out to memory
          bclr #23,a1           ;clear the highest bit which was already
                                ;shifted out by the extra shift
        .endl
        rts             ;modifies a,b,x, r1/m1 is advanced by n1

FormatFreq_sddd         ;x0 = frequency (X:RxCarFreq), Y:(r1)/m1 where to format
        move #SampleFreq/(2.0*DecimateRatio)/1000.0/2,x1
        mpy x0,x1,a
        asl a #<3,n1
        jsr <FormatSign         ;format the sign and then three digits
        jsr <FormatFraction
;        jsr <FormatDot          ;format the decimal dot
;        move #<2,n1             ;two more digits after the dot
;        jsr <FormatFraction
        rts             ;modifies a,b,x; r1 is advanced by 1+3

FormatFreq_sdd          ;x0 = frequency (X:RxCarFreq), Y:(r1)/m1 where to format
        move #SampleFreq/(2.0*DecimateRatio)/1000.0/2,x1
        mpy x0,x1,a
        asl a #<1,n1
        jsr <FormatSign         ;format the sign
        jsr <FormatFraction
        move (r1)-
        move #<2,n1
        jsr <FormatFraction
        rts             ;modifies a,b,x; r1 is advanced by 1+2

FormatPosInt_dddd       ;x0 = positive integer, Y:(r1)/m1 where to format
        clr a #>1.0/10000.0,x1
        mpy x0,x1,b
        move b0,a1
        asr a #<4,n1    ;format the integer as four digits
        jsr <FormatFraction
        rts             ;modifies a,b,x; r1 is advanced by 4

FormatLog2_sddd_dB      ;a = 2-based logarythm (computed by Log2)
        rep #6          ;first scale down by 64 so it is below 1.0
          asr a
        move a,x0
FormatLog2frac_sddd_dB
        move #>@log(2)/@log(10)/1000.0*10.0*64.0,x1
        mpy x0,x1,a #<3,n1
        jsr <FormatSign
        jsr <FormatFraction
        rts

FormatPosLog2frac_ddd_dB
        move #>@log(2)/@log(10)/1000.0*10.0*64.0,x1
        mpy x0,x1,a #<3,n1
        jsr <FormatFraction
        rts

FormatPosLog2frac_dd_dB
        move #>@log(2)/@log(10)/1000.0*10.0*64.0,x1
        mpy x0,x1,a #<1,n1
        jsr <FormatFraction
        move (r1)-
        move #<2,n1
        jsr <FormatFraction
        rts

Log2                    ;a2:1:0 = 56-bit argument (0.0,+256.0)
	tst a #<0,r0    ;r0 ready for finding the integer part
	jmi <Log2_Neg   ;if argument negative...
	jeq <Log2_Inf   ;if zero...
	jnr <Log2_Frac  ;if already normalized
Log2_Norm               ;normalize the argument to the range <0.5,1.0)
	rep #8
	  norm r0,a     ;r0 counts the number of shifts performed
	jnn <Log2_Norm
			;r0 = approx. integer result
Log2_Frac               ;a = argument scaled to range <0.5,1.0) = X
	clr a a,x0      ;x0 = X
	move r0,a2      ;now we compute the finer fractional approximation
	asr a x0,x1     ;a = approx. integer part
	bchg #23,x0     ;x0 = X-1
	add x0,a        ;a += 2*(X-1) => first, linear correction
	add x0,a
	bchg #22,x1     ;x1 = X-1/2
	mpyr x0,x1,b #0.371,x1 ;b = (X-1)*(X-1/2)
	sub b,a b,x0    ;a += (X-1)*(X-1/2)
	macr -x0,x1,a   ;a += 0.371*(X-1)*(X-1/2)
	rts             ;a = log2(input) <-48.0,+8.0) => second, parabolic corr.
			;b,x,r0 are modified
			;error RMS is 0.005, maximum error is 0.008
Log2_Neg                ;for zero and negative arguments we give the most
Log2_Inf clr a          ;negative result: a = -48.0
	move #<$E8,a2
	rts             ;execution time is about 80 clock cycles
			;but becomes larger for very small arguments

      endif ;of RxReport
;------------------------------------------------------------------------------
Initialize      ;initialize registers, buffers, windows, etc.

        ori #%00000100,omr      ;enable the sine ROM table
                                ;this is for the IQ routine

        clr a #WindowInpTap,r0  ;clean the window taps
        move #WindowLen-1,m0
        rep #WindowLen
          move a,L:(r0)+
        move #WindowOutTap,r0
        rep #WindowLen
          move a,L:(r0)+

      if WindowLen<=256
        move #$100,r0           ;initialize the sine/cosine table
        move #<$FF,m0           ;by taking every Nth value from the on-chip
        move #<$100/WindowLen,n0 ;ROM sine table (N=256/WindowLen)
        move #SinTable,r4         ;This works only for FFT lengths up to 256
        move #<WindowLen-1,m4     ;note that the table for the FFT starts
        move #<3*(WindowLen/4),n4 ;with cos=-1.0 and sin=0.0
        .loop #WindowLen        ;so we make a dedicated entry: FFTcoeff
          move Y:(r0)+n0,a      ;in the SinTable
          move a,X:(r4+n4)      ;X memory contains the cosine
          move a,Y:(r4)+        ;Y memory contains the sine
        .endl
      else
        move #SinTable,r4       ;initialize the sine/cosine coeff. table
        move #WindowLen-1,m4    ;we start at phase = 0
        clr a #2.0/WindowLen,b  ;and we step by 2*PI/WindowLen
        move ab,L:<Scratch      ;a = phase, b = step
        .loop #WindowLen
          move a1,x0
          jsr <IQ               ;a = the cosine => X
          move ab,L:(r4)+       ;b = the sine => Y
          move L:<Scratch,ab
          add b,a
          move a1,X:<Scratch
        .endl
      endif
                                ;r4/m4 points to the SinTable
      clr a #TxInitVectors,r5   ;one way to set the initial vectors
      move #<DataCarriers-1,m5
      move #>8,b                ;you may load b with 1..(WindowLen/DataCarriers)
      .loop #DataCarriers
        add b,a L:(r4),x
        move a,n4
        move x,L:(r5)+
        move (r4)+n4
      .endl

      move #InterleavePattern,r0        ;read the interleave definition
      move #<DataCarriers-1,m0          ;and prepare the pattern for the Tx
      move #TxIntlvPatt,r1
      move m0,m1
      move #>WordsPerBlock,x1
      .loop #DataCarriers
        move Y:(r0)+,x0
        mpy x0,x1,a
        asr a
        move a0,Y:(r1)+
      .endl

      move #>RxDataSampleLen,x1
      move #>1,y1
      move #RxIntlvPatt0,r1
      .loop #DataCarriers/2
        move Y:(r0)+,x0
        mpy x0,x1,a
        move a0,a
        sub x1,a Y:(r0)+,x0
        mpy x0,x1,a a,Y:(r1)+
        move a0,a
        add x1,a
        sub y1,a
        move a,Y:(r1)+
      .endl

      move #RxIntlvPatt1,r1
      .loop #DataCarriers/2
        move Y:(r0)+,x0
        mpy x0,x1,a
        move a0,a
        sub x1,a
        sub y1,a Y:(r0)+,x0
        mpy x0,x1,a a,Y:(r1)+
        move a0,a
        add x1,a
        move a,Y:(r1)+
      .endl

      if NewBaudRate        ;set new baud rate (copied from LEONID)
        movep #(xtal+2*16*NewBaudRate)/(2*2*16*NewBaudRate)-1,X:<<$FFF2
      endif

        move #CodecBuff+2,r7    ;for the CODEC's interrupt routine
        move #CodecBuffLen*4-1,m7

        move #CodecBuff,r2      ;for "cdctrl" to initialize the buffer
        move #<4-1,n2
        move #CodecBuffLen*4-1,m2
                        ;initialize input/output control words in the buffer
                        ;zero input/output data
      if EVM56K         ;for EVM56002 use MIC input
        ctrlcd  1,r2,CodecBuffLen,MIC,RxGain,RxGain,LINEO|HEADP,TxAttenuate,TxAttenuate
      else              ;for DSPCARD4 use LINE input
        ctrlcd  1,r2,CodecBuffLen,LINEI,RxGain,RxGain,LINEO|HEADP,TxAttenuate,TxAttenuate
      endif
        opencd SampleFreq/1000.0,HPF     ;start taking samples at given rate

        jmp <ProcessLoop

      if SPY

SpySync move a10,L:<SpySave     ;output: carry=1 => no spy request
        move a2,X:<SpySave+1    ;carry=0 => spy request !
        move x0,Y:<SpySave+1    ;SPY_BlockLen words (jsr <SpyA) should follow
        move x1,Y:<SpyCount
        move X:<SpyCount,a
        tst a
        jne <SpyCont
        lookc 0
        jcs <Spy_end
        move #>'S',a
        cmp x0,a
        ori #$01,ccr
        jne <Spy_end
        move #>'P',x0
        putc
        move #>SPY_BlockLen,a
        move a,X:<SpyCount
SpyCont andi #$FE,ccr
        jmp <Spy_end

SpyA    move a10,L:<SpySave
        move a2,X:<SpySave+1
        move x0,Y:<SpySave+1
        move x1,Y:<SpyCount
        move X:<SpyCount,a
        tst a
        jne <Spy_copy

Spy_check
        lookc 0
        jcs <Spy_end
        move #>'S',a
        cmp x0,a
        jne <Spy_end
        move #>'P',x0
        putc
        move #>SPY_BlockLen,a
Spy_copy
        move #>1,x0
        sub x0,a
        move a,X:<SpyCount

        move X:<SpySave,a
        rep #8
          lsr a
        move a1,x0
	putc
        move X:<SpySave,a
        rep #16
          lsr a
        move a1,x0
        putc

Spy_end move L:<SpySave,a10
        move X:<SpySave+1,a2
        move Y:<SpySave+1,x0
        move Y:<SpyCount,x1
        rts

      endif ; of SPY

;*************************************************************************
;Internal data RAM

        LOMEM X:$0000,Y:$0000,L:$0000
        HIMEM X:$00FF,Y:$00FF,L:$00FF

        org L:user_data

        if SPY
SpySave dc 0,0
SpyCount dc 0
        endif

RxTmpI  ds 1    ;temporary storages for the receiver's symbol decoder
RxTmpQ  ds 1

Scratch ds 1    ;general purpose scratch variable

Rand    dc $954820584783        ;seed and increment for the random
RandInc dc $894659236475        ;number generator

LastL = *
        org X:LastL
        org Y:LastL

        org X:
WindowInpTapPtr dc WindowInpTap
WindowOutTapPtr dc WindowOutTap

RxCodecPtr    dc CodecBuff
TxCodecPtr    dc CodecBuff+4*31 ;-4*3*4

RxSliceSeparSamp  dc RxSliceSepar*DecimateRatio
RxSliceSeparDeci  dc RxSliceSepar

RxCarPhase    dc 0      ;Rx mixer (doppler corrector) carrier phase
RxCarFreq     dc 0      ;and initial frequency

RxFFTphaseCorr dc 0
RxPipePtr     dc 0
RxPhaseDevPtr dc 0
RxDecodeDevPtr dc 0

RxDataPtr       dc 0
RxSymbolPhase   dc 0    ;0,1,2,3 = which symbol slice is going now
RxDataFirstBin  dc 0    ;first FFT bin to be taken in data phase

TxStateRoutine  dc TxStartIdle  ;label to be jumped to in particular state
TxStateCounter  dc 0    

TxState         dc 0    ;various state bits
TxSta_Symbol     equ 0
TxSta_SkipSlice  equ 1
TxSta_PTTonReq   equ 2 ;definition of bits in the TxState
TxSta_PTToffReq  equ 3 ;
TxSta_PTTabortReq equ 4

TxIntlvPtr      dc 0
TxMsgPtr        dc 0
        if !SPY
TxCharBuffRd    dc TxCharBuff
TxCharBuffWr    dc TxCharBuff
        endif

RxStateRoutine  dc RxStartIdle
RxStateCounter  dc 0
RxAcceptCount   dc 0
RxTimeCorrHold  dc 0
RxDataSamplePtr dc 0

RxState         dc 0
RxSta_DCD       equ 2   ;definition of bits in the RxState
RxSta_NowDCD    equ 3
RxSta_AbortReq  equ 4
RxSta_QualityReq equ 5

RxSignalMinDev   dc 0
RxDecodeFirstBin dc 0
RxDecodeBestCarr dc 0
RxDecodeMinNoise dc 0

        if RxAGC
RxAudioPeak     dc 0            ;peak level of the input audio
RxAudioMS       dc 0            ;Mean Square (the power) of the input audio
RxAGCcount      dc RxAGCfall
        endif

        org Y:

;some Y-storage internal variables

TxDataBlock     dsm WordsPerBlock       ;data block to be sent out

        org X:
LastX = *
        org Y:
LastY = *
        if @cvi(LastX)>=@cvi(LastY)
          org L:LastX
        else
          org L:LastY
        endif

TxWorkVectors  dsm DataCarriers         ;vectors we are actually
                                        ;working with when transmitting
TxVect         dsm DataCarriers         ;vector of symbols to be transmitted
                                        ;this will be sent out in two parts
                                        ;first even elements then odd ones
                                        ;to shift the symbols on adjacent
                                        ;carriers by SymbolSepar/2

;*************************************************************************
;External L (X/Y) data RAM

        if EVM56K
          LOMEM X:$2000,Y:$2000,L:$2000
          HIMEM X:$3FFF,Y:$3FFF,L:$3FFF
        else
          LOMEM X:$0200,Y:$0200,L:$0200
          HIMEM X:$1FFF,Y:$1FFF,L:$1FFF
        endif

      if EVM56K
        org L:$2000
      else
        org L:$200
      endif

        org L:
FFTbuff dsm WindowLen   ;Work buffer for the FFT routine

WindowInpTap dsm WindowLen
WindowOutTap dsm WindowLen

SinTable dsm WindowLen   ;sine/cosine table (full period)
FFTcoef equ SinTable+WindowLen/2 ;an entry for the FFT sine/cosine coeff.

TxInitVectors  dsm DataCarriers         ;initial vectors (phases)

;for the sync. phase
RxSyncPhaseDev dsm 4*RxSyncPhaseDevLen
RxSyncPhaseDevMid dsm 4*RxSyncPhaseDevLen

;for the data phase
RxDataPhaseDev dsm 4*RxDataPhaseDevLen
RxDataPhaseDevMid dsm 4*RxDataPhaseDevLen
RxDecodeDev dsm 2*2*RxDataScanMargin
RxDecodeDevMid dsm 2*2*RxDataScanMargin

LastL = *
        org X:LastL
        org Y:LastL

        org Y:
AliasFilterInpI dsm AliasFilterLen      ;anti-alias input filter (I/Q)
AliasFilterInpQ dsm AliasFilterLen

        org X:
AliasFilterOutI dsm AliasFilterLen      ;anti-alias output filter (I/Q)
AliasFilterOutQ dsm AliasFilterLen

        org X:
LastX = *
        org Y:
LastY = *
        if @cvi(LastX)>=@cvi(LastY)
          org L:LastX
        else
          org L:LastY
        endif

CodecBuff dsm CodecBuffLen*4    ;CODEC's input/output buffer

;*************************************************************************
;external Y-only data RAM

        if EVM56K
          LOMEM Y:$0200
          HIMEM Y:$1FFF
        else
          LOMEM Y:$2000
          HIMEM Y:$3FFF
        endif

      if EVM56K
        org Y:$200
      else
        org Y:$2000
      endif

        org Y:
DataWindowInp   dsm WindowLen           ;input/output FFT window
DataWindowOut   dsm WindowLen           ;for data symbols Tx/Rx

;the following we need for sync. phase
RxSyncPhaseBuff dsm RxSyncScanLen    ;there we put differential phases
                                     ;we compute at each FFT slice
RxSyncPhasePipe dsm 4*RxSyncScanLen  ;the pipe to buffer the phases
                                     ;during the Rx sync. phase

;the following we need for the data phase
RxDataPhasePipe dsm 4*RxDataScanLen     ;FFT phase pipe for later differenciation
RxDataPhaseBuff dsm RxDataScanLen       ;here we store differential phase
RxDataSamplePipe dsm DataInterleave*2*RxDataSampleLen ;soft decision and de-interleave pipe

;Transmitter interleave pipe
TxIntlvPipe     dsm WordsPerBlock*DataInterleave
;Transmitter interleave pattern
TxIntlvPatt     dsm DataCarriers

;Receiver interleave pattern
RxIntlvPatt0    dsm DataCarriers
RxIntlvPatt1    dsm DataCarriers

;Primary definition of the interleave pattern
InterleavePattern dsm DataCarriers

;FHT buffer
FHTbuff dsm BitsPerBlock

TxStartCharSeq dsm TxSyncLen
TxStopCharSeq  dsm TxStopLen+DataInterleave+1

        if !SPY
TxCharBuff dsm TxCharBuffLen
        endif

        if (!SPY)&&TxReport
TxStartMsg dc CR,LF,'Transmitter ON ...',0
TxStopMsg  dc 'Transmitter OFF.',CR,LF,0
TxAbortMsg dc CR,LF,'Transmitter emergency shutdown',CR,LF,0
        endif

        if (!SPY)&&RxReport
RxLockMsg      dc CR,LF,'Signal acquired at '
RxLockMsg_freq dc '+00 Hz S/N = '
RxLockMsg_SNR  dc '000 dB',CR,LF,0
RxLockLostMsg  dc CR,LF,'Signal lost',CR,LF,0
RxLockAbortMsg dc CR,LF,'Signal aborted',CR,LF,0
RxQualMsg      dc 'Signal S/N = '
RxQualMsg_SNR  dc '000 dB FEC S/N = '
RxQualMsg_FEC  dc '000 dB Freq. offset = '
RxQualMsg_freq dc '+00 Hz',CR,LF,0
        endif   ;of RxReport



;*************************************************************************
;constant tables: FIR and window shapes, FFT coefficiants, etc.

        if EVM56K
          LOMEM X:$2000,Y:$0200,L:$2000
          HIMEM X:$3FFF,Y:$3FFF,L:$3FFF
        else
          LOMEM X:$0200,Y:$0200,L:$0200
          HIMEM X:$1FFF,Y:$3FFF,L:$1FFF
        endif

        nolist
        include 'alias.dat'
        include 'symbol.dat'
        list

        org Y:TxStartCharSeq     ;TxSyncLen characters
        dc CR,'*START* 64 BPSK tones, 10 baud *START*',CR

        org Y:TxStopCharSeq
        dc CR,'*EOT* *EOT* *EOT*',CR       ;TxStopLen characters
        dc 0,0,0,0,0,0,0,0      ;DataInterleave+1 characters
        dc 0,0,0,0,0,0,0,0
        dc 0,0,0,0,0,0,0,0
        dc 0,0,0,0,0,0,0,0
        dc 0

;Here we define the interleave pattern
        org Y:InterleavePattern
        dc 4,5,6,7
        dc 4,5,6,7
        dc 4,5,6,7
        dc 4,5,6,7
        dc 4,5,6,7
        dc 4,5,6,7
        dc 4,5,6,7
        dc 4,5,6,7
        dc 4,5,6,7
        dc 4,5,6,7
        dc 4,5,6,7
        dc 4,5,6,7
        dc 4,5,6,7
        dc 4,5,6,7
        dc 4,5,6,7
        dc 4,5,6,7

      if 0
        dc 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
        dc 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
        dc 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
        dc 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
      endif

      if 0
        dc 1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16
        dc 17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32
        dc 31,30,29,28,27,26,25,24,23,22,21,20,19,18,16
        dc 15,14,13,12,11,10,9,8,7,6,5,4,3,2,1,0
      endif

;*************************************************************************

        end

